
#ifndef SIM_HEADER
#define SIM_HEADER



#include "tensor_runtime.h"
#include "tensor_utils.h"
#include "debug.h"
#include "profiling.h"
#include "fp16_conversion.h"
#include "global_data.h"
#include "error.h"
#include "tensor.h"
#include "op_overheads.h"
#include "half_precision_api.h"
#include "approx_utils.h"
#include "device_math.h"
#include "global_data.h"
#include "approx_knob_utils.h"


#include <unordered_map>
#include <sstream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
#include <cassert>




//N is new_data's size
//n, c, h, w are the dimensions of new_data
__global__
void postInterpolateRow(int N, int n, int c, int h, int w,
			float* data, int int_row){

  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  for(int i = index; i < N; i += stride){
    int col = ((i % (c * h * w)) % (h * w)) % w;
    int row = ((i % (c * h * w)) % (h * w)) / w;
    int ch = (i % (c * h * w)) / (h * w);
    int n = i / (c * h * w);

    if((row % int_row == 1) && (row != 0) && (row != h-1))
      data[n * (c * h * w) + ch * (h * w) + row * (w) + col] =
	(data[n * (c * h * w) + ch * (h * w) + (row - 1) * (w) + col] +
	 data[n * (c * h * w) + ch * (h * w) + (row + 1)  * (w) + col]) / 2;

  }
}



__global__
void postInterpolateCol(int N, int n, int c, int h, int w,
			float* data, int int_col){

  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  for(int i = index; i < N; i += stride){
    int col = ((i % (c * h * w)) % (h * w)) % w;
    int row = ((i % (c * h * w)) % (h * w)) / w;
    int ch = (i % (c * h * w)) / (h * w);
    int n = i / (c * h * w);

    if((col % int_col == 1) && (col != 0) && (col != w-1))
      data[n * (c * h * w) + ch * (h * w) + row * (w) + col] =
	(data[n * (c * h * w) + ch * (h * w) + row * (w) + (col-1) ] +
	 data[n * (c * h * w) + ch * (h * w) + row * (w) + (col+1) ])/2;

  }
}




// A 'Simulation' of perforated tensor convolution
void* tensorConvPerfSim(void* input_ptr, void* filter_ptr,
			int vertical_pad, int horizontal_pad,
			int vertical_stride, int horizontal_stride,
			int conv_mode, int conv_groups,
			int row, int col){
  

  INFO("*** TensorConvolution \n");
  profileEvent("tensorConv");

  Tensor* input = (Tensor*) input_ptr;
  Tensor* filter = (Tensor*) filter_ptr;

  cudnnConvolutionDescriptor_t convDesc;
  cudnnConvolutionFwdAlgo_t convAlgo;
  cudnnConvolutionMode_t mode;
  
  if(conv_mode == 0)
    mode = CUDNN_CONVOLUTION;
  else if(conv_mode == 1)
    mode = CUDNN_CROSS_CORRELATION;

  float alpha = 1.0f, beta = 0.0f;

  hostToDeviceCopy(input);
  hostToDeviceCopy(filter);

  INFO("vertical_stride = %lu, horizontal_stride = %lu \n",
       vertical_stride, horizontal_stride);

  checkCUDNN(cudnnCreateConvolutionDescriptor(&convDesc));

  //FIXME: Current hack to preserve backward compatibilty
  if(conv_groups == 0){
    conv_groups = 1;
  }

  // NOTE: Adding support for grouped convolution
  checkCUDNN(cudnnSetConvolutionGroupCount(convDesc, conv_groups));

  int new_v = vertical_stride + 0;
  int new_h = horizontal_stride + 0;
  cudnnDataType_t computeType = CUDNN_DATA_FLOAT;
  
  checkCUDNN(cudnnSetConvolution2dDescriptor(convDesc,
					     vertical_pad, horizontal_pad, // conv padding
					     new_v, new_h, // conv strides
					     1, 1, // upscaling values
					     mode , // mode is configurable
					     computeType)); // defines compute precision

  int n, c, h, w; // output dimensions
  // Find dimension of convolution output
  checkCUDNN(cudnnGetConvolution2dForwardOutputDim(convDesc,
						   input->tensor_desc,
						   filter->filter_desc,
						   &n, &c, &h, &w));


  DEBUG("**Output Tensor Dims, n = %d, c = %d, h = %d, w = %d \n", n, c, h, w);

  Tensor* output;
  if(input->data_format == CUDNN_TENSOR_NCHW)
    output = (Tensor*) create4DTensor((cudnnDataType_t) input->data_type,
				      CUDNN_TENSOR_NCHW, n, c, h, w);
  else if(input->data_format == CUDNN_TENSOR_NHWC){
    DEBUG("* NHWC Format \n");
    output = (Tensor*) create4DTensor((cudnnDataType_t) input->data_type,
				      CUDNN_TENSOR_NHWC, n, h, w, c);
  }
  else
    ERROR("Unsupported Tensor Type");

  // NOTE: Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE);
  // NOTE: Necessary to insert the above call for every output tensor

  DEBUG("tensor->data_type = %d, tensor->data_format = %d, N = %d, C = %d, H = %d, W = %d \n",
	output->data_type, output->data_format, output->dims.dim_sizes[0],
	output->dims.dim_sizes[1],
	output->dims.dim_sizes[2], output->dims.dim_sizes[3]);

  if(convDesc == NULL || input->tensor_desc == NULL ||
     filter->filter_desc == NULL || output->tensor_desc == NULL)
    ERROR("NULL descriptor! \n");



  // NOTE-FIXIT: function failing for NHWC formats - perhaps some CUDNN support is lacking
  checkCUDNN(cudnnGetConvolutionForwardAlgorithm(cudnnHandle,
						 input->tensor_desc,
						 filter->filter_desc,
						 convDesc,
						 output->tensor_desc,
						 CUDNN_CONVOLUTION_FWD_PREFER_FASTEST,
						 //CUDNN_CONVOLUTION_FWD_NO_WORKSPACE,
						 0,
						 &convAlgo));


  DEBUG("ConvAlgo = %d, FFT = %d, GEMM = %d, WINOGRAD = %d \n", convAlgo,
	CUDNN_CONVOLUTION_FWD_ALGO_FFT, CUDNN_CONVOLUTION_FWD_ALGO_GEMM,
	CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD);


  // FIXIT: Algo shouldn't be hardcoded
  convAlgo = CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM;

  size_t workspace_size;
  checkCUDNN(cudnnGetConvolutionForwardWorkspaceSize(cudnnHandle,
						     input->tensor_desc,
						     filter->filter_desc,
						     convDesc,
						     output->tensor_desc,
						     convAlgo,
						     &workspace_size));

  // Allocating memory for the convolution workspace
  void* workspace;
  checkCudaErrors(cudaMalloc(&workspace, workspace_size));
  DEBUG("workspace size = %d \n", workspace_size);


  checkCUDNN(cudnnConvolutionForward(cudnnHandle, &alpha, input->tensor_desc,
				     input->gpu_data, filter->filter_desc, filter->gpu_data,
				     convDesc, convAlgo, workspace, workspace_size,
				     &beta, output->tensor_desc, output->gpu_data));


  h = (2 * vertical_pad + input->dims.dim_sizes[2] -
       filter->dims.dim_sizes[2]) / vertical_stride + 1;
  
  w = (2 * horizontal_pad + input->dims.dim_sizes[3] -
       filter->dims.dim_sizes[3]) / horizontal_stride + 1;


  int numBlocks = (n * c * h * w  + 127) / 128;

  if (row > 0)
    postInterpolateRow<<<numBlocks,128>>>(n * c * h * w, n, c, h, w,
				         (float *) output->gpu_data, row);

  if (col > 0)
    postInterpolateCol<<<numBlocks,128>>>(n * c * h * w, n, c, h, w,
				         (float *) output->gpu_data, col);


  profileEvent("tensorConv_end", true);

  return output;
}





//N is new_data's size
//n, c, h, w are the dimensions of new_data
__global__
void sampleFilterElems(int N,
		       int n, int c, int h, int w,
		       float* data,
		       int skip_elem, int skip_offset,
		       float mul_factor,
		       float* newData){

  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  
  for(int i = index; i < N; i += stride){
    int col = ((i % (c * h * w)) % (h * w)) % w;
    int row = ((i % (c * h * w)) % (h * w)) / w;
    int ch = (i % (c * h * w)) / (h * w);
    int n = i / (c * h * w);

    int local_index = (ch * (h * w)) + (row * w) + col;

    if(skip_elem == 3 && h == 3 && w == 3){
      skip_offset = (skip_offset + ch) % w; // wrap around skip offsets
    }

    if(local_index % skip_elem  == skip_offset)
       newData[n * (c * h * w) + ch * (h * w) + row * (w) + col] = 0;
    else
      newData[n * (c * h * w) + ch * (h * w) + row * (w) + col] =
      data[n * (c * h * w) + ch * (h * w) + row * (w) + col] * mul_factor;
      
  }
}





void sampleFilter(Tensor* newFilter, Tensor* filter,
		  int skip_rate, int skip_offset){

  int n = filter->dims.dim_sizes[0];
  int c = filter->dims.dim_sizes[1];
  int h = filter->dims.dim_sizes[2];
  int w = filter->dims.dim_sizes[3];
    
  int numBlocks = (n * c * h * w  + 127) / 128;
  int N = n * c * h * w;

  float mul_factor = (skip_rate * 1.0) / (skip_rate - 1); 

  //float mul_factor = (skip_rate * 1.0) / (skip_rate - 1);
  //mul_factor = (mul_factor + 1.0) / 2;

  
  DEBUG ("mul_factor = %f \n", mul_factor);

  
  sampleFilterElems<<<numBlocks,128>>>(N,
				       n, c, h, w,
				       (float *) filter->gpu_data,
				       skip_rate, skip_offset, mul_factor,
				       (float *) newFilter->gpu_data);

}



// A 'Simulation' of perforated tensor convolution
void* tensorConvSampSim(void* input_ptr, void* filter_ptr,
			int vertical_pad, int horizontal_pad,
			int vertical_stride, int horizontal_stride,
			int conv_mode, int conv_groups,
			int skip_rate, int skip_offset){
  

  INFO("*** TensorConvolution \n");
  profileEvent("tensorConv");

  Tensor* input = (Tensor*) input_ptr;
  Tensor* filter = (Tensor*) filter_ptr;

  
  cudnnConvolutionDescriptor_t convDesc;
  cudnnConvolutionFwdAlgo_t convAlgo;  
  cudnnConvolutionMode_t mode;
  
  if(conv_mode == 0)
    mode = CUDNN_CONVOLUTION;
  else if(conv_mode == 1)
    mode = CUDNN_CROSS_CORRELATION;

  float alpha = 1.0f, beta = 0.0f;

  hostToDeviceCopy(input);
  hostToDeviceCopy(filter);

  convertToFP32(input);
  convertToFP32(filter);

  Tensor* newFilter;
  newFilter = (Tensor *) create4DTensor((cudnnDataType_t) float_type,
					CUDNN_TENSOR_NCHW, filter->dims.dim_sizes[0],
					filter->dims.dim_sizes[1], filter->dims.dim_sizes[2],
					filter->dims.dim_sizes[3]);


  // Zeroing (+Scaling) Filter elements to 'Simulate' input sampling
  sampleFilter(newFilter, filter, skip_rate, skip_offset);
  

  INFO("vertical_stride = %lu, horizontal_stride = %lu \n",
       vertical_stride, horizontal_stride);

  checkCUDNN(cudnnCreateConvolutionDescriptor(&convDesc));

  //FIXME: Current hack to preserve backward compatibilty
  if(conv_groups == 0){
    conv_groups = 1;
  }

  // NOTE: Adding support for grouped convolution
  checkCUDNN(cudnnSetConvolutionGroupCount(convDesc, conv_groups));

  int new_v = vertical_stride + 0;
  int new_h = horizontal_stride + 0;
  cudnnDataType_t computeType = CUDNN_DATA_FLOAT;
  
  checkCUDNN(cudnnSetConvolution2dDescriptor(convDesc,
					     vertical_pad, horizontal_pad, // conv padding
					     new_v, new_h, // conv strides
					     1, 1, // upscaling values
					     mode , // mode is configurable
					     computeType)); // defines compute precision

  int n, c, h, w; // output dimensions
  // Find dimension of convolution output
  checkCUDNN(cudnnGetConvolution2dForwardOutputDim(convDesc,
						   input->tensor_desc,
						   filter->filter_desc,
						   &n, &c, &h, &w));


  DEBUG("**Output Tensor Dims, n = %d, c = %d, h = %d, w = %d \n", n, c, h, w);

  Tensor* output;
  output = (Tensor*) create4DTensor((cudnnDataType_t) float_type, 
				      CUDNN_TENSOR_NCHW, n, c, h, w);
  

  // NOTE: Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE);
  // NOTE: Necessary to insert the above call for every output tensor

  DEBUG("tensor->data_type = %d, tensor->data_format = %d, N = %d, C = %d, H = %d, W = %d \n",
	output->data_type, output->data_format, output->dims.dim_sizes[0],
	output->dims.dim_sizes[1],
	output->dims.dim_sizes[2], output->dims.dim_sizes[3]);

  if(convDesc == NULL || input->tensor_desc == NULL ||
     filter->filter_desc == NULL || output->tensor_desc == NULL)
    ERROR("NULL descriptor! \n");


  // NOTE-FIXIT: function failing for NHWC formats - perhaps some CUDNN support is lacking
  checkCUDNN(cudnnGetConvolutionForwardAlgorithm(cudnnHandle,
						 input->tensor_desc,
						 filter->filter_desc,
						 convDesc,
						 output->tensor_desc,
						 CUDNN_CONVOLUTION_FWD_PREFER_FASTEST,
						 //CUDNN_CONVOLUTION_FWD_NO_WORKSPACE,
						 0,
						 &convAlgo));


  DEBUG("ConvAlgo = %d, FFT = %d, GEMM = %d, WINOGRAD = %d \n", convAlgo,
	CUDNN_CONVOLUTION_FWD_ALGO_FFT, CUDNN_CONVOLUTION_FWD_ALGO_GEMM,
	CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD);


  // NOTE: Using GEMM-based Algo
  convAlgo = CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM;

  size_t workspace_size;
  checkCUDNN(cudnnGetConvolutionForwardWorkspaceSize(cudnnHandle,
						     input->tensor_desc,
						     filter->filter_desc,
						     convDesc,
						     output->tensor_desc,
						     convAlgo,
						     &workspace_size));

  // Allocating memory for the convolution workspace
  void* workspace;
  checkCudaErrors(cudaMalloc(&workspace, workspace_size));
  DEBUG("workspace size = %d \n", workspace_size);


  checkCUDNN(cudnnConvolutionForward(cudnnHandle, &alpha, input->tensor_desc,
				     input->gpu_data, filter->filter_desc, newFilter->gpu_data,
				     convDesc, convAlgo, workspace, workspace_size,
				     &beta, output->tensor_desc, output->gpu_data));


 
  freeTensor(newFilter);
  profileEvent("tensorConv_end", true);

  return output;
}










void sampleFilter2(Tensor* newFilter, Tensor* filter,
		   int skip_rate, int skip_offset, float interpolation_rate){

  int n = filter->dims.dim_sizes[0];
  int c = filter->dims.dim_sizes[1];
  int h = filter->dims.dim_sizes[2];
  int w = filter->dims.dim_sizes[3];
    
  int numBlocks = (n * c * h * w  + 127) / 128;
  int N = n * c * h * w;

  float mul_factor;
  mul_factor = (skip_rate * 1.0) / (skip_rate - 1);
  mul_factor = 1 + (interpolation_rate * (mul_factor - 1.0));
  DEBUG ("mul_factor = %f \n", mul_factor);
  
  sampleFilterElems<<<numBlocks,128>>>(N,
				       n, c, h, w,
				       (float *) filter->gpu_data,
				       skip_rate, skip_offset, mul_factor,
				       (float *) newFilter->gpu_data);
}



// A 'Simulation' of perforated tensor convolution
void* tensorConvSampSim2(void* input_ptr, void* filter_ptr,
			 int vertical_pad, int horizontal_pad,
			 int vertical_stride, int horizontal_stride,
			 int conv_mode, int conv_groups,
			 int skip_rate, int skip_offset, float interpolation_rate){
  

  INFO("*** TensorConvolution \n");
  profileEvent("tensorConv");

  Tensor* input = (Tensor*) input_ptr;
  Tensor* filter = (Tensor*) filter_ptr;

  
  cudnnConvolutionDescriptor_t convDesc;
  cudnnConvolutionFwdAlgo_t convAlgo;  
  cudnnConvolutionMode_t mode;
  
  if(conv_mode == 0)
    mode = CUDNN_CONVOLUTION;
  else if(conv_mode == 1)
    mode = CUDNN_CROSS_CORRELATION;

  float alpha = 1.0f, beta = 0.0f;

  hostToDeviceCopy(input);
  hostToDeviceCopy(filter);

  convertToFP32(input);
  convertToFP32(filter);

  Tensor* newFilter;
  newFilter = (Tensor *) create4DTensor((cudnnDataType_t) float_type,
					CUDNN_TENSOR_NCHW, filter->dims.dim_sizes[0],
					filter->dims.dim_sizes[1], filter->dims.dim_sizes[2],
					filter->dims.dim_sizes[3]);


  // Zeroing (+Scaling) Filter elements to 'Simulate' input sampling
  sampleFilter2(newFilter, filter, skip_rate, skip_offset, interpolation_rate);
  

  INFO("vertical_stride = %lu, horizontal_stride = %lu \n",
       vertical_stride, horizontal_stride);

  checkCUDNN(cudnnCreateConvolutionDescriptor(&convDesc));

  //FIXME: Current hack to preserve backward compatibilty
  if(conv_groups == 0){
    conv_groups = 1;
  }

  // NOTE: Adding support for grouped convolution
  checkCUDNN(cudnnSetConvolutionGroupCount(convDesc, conv_groups));

  int new_v = vertical_stride + 0;
  int new_h = horizontal_stride + 0;
  cudnnDataType_t computeType = CUDNN_DATA_FLOAT;
  
  checkCUDNN(cudnnSetConvolution2dDescriptor(convDesc,
					     vertical_pad, horizontal_pad, // conv padding
					     new_v, new_h, // conv strides
					     1, 1, // upscaling values
					     mode , // mode is configurable
					     computeType)); // defines compute precision

  int n, c, h, w; // output dimensions
  // Find dimension of convolution output
  checkCUDNN(cudnnGetConvolution2dForwardOutputDim(convDesc,
						   input->tensor_desc,
						   filter->filter_desc,
						   &n, &c, &h, &w));


  DEBUG("**Output Tensor Dims, n = %d, c = %d, h = %d, w = %d \n", n, c, h, w);

  Tensor* output;
  output = (Tensor*) create4DTensor((cudnnDataType_t) float_type, 
				      CUDNN_TENSOR_NCHW, n, c, h, w);
  

  // NOTE: Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE);
  // NOTE: Necessary to insert the above call for every output tensor

  DEBUG("tensor->data_type = %d, tensor->data_format = %d, N = %d, C = %d, H = %d, W = %d \n",
	output->data_type, output->data_format, output->dims.dim_sizes[0],
	output->dims.dim_sizes[1],
	output->dims.dim_sizes[2], output->dims.dim_sizes[3]);

  if(convDesc == NULL || input->tensor_desc == NULL ||
     filter->filter_desc == NULL || output->tensor_desc == NULL)
    ERROR("NULL descriptor! \n");


  // NOTE-FIXIT: function failing for NHWC formats - perhaps some CUDNN support is lacking
  checkCUDNN(cudnnGetConvolutionForwardAlgorithm(cudnnHandle,
						 input->tensor_desc,
						 filter->filter_desc,
						 convDesc,
						 output->tensor_desc,
						 CUDNN_CONVOLUTION_FWD_PREFER_FASTEST,
						 //CUDNN_CONVOLUTION_FWD_NO_WORKSPACE,
						 0,
						 &convAlgo));


  DEBUG("ConvAlgo = %d, FFT = %d, GEMM = %d, WINOGRAD = %d \n", convAlgo,
	CUDNN_CONVOLUTION_FWD_ALGO_FFT, CUDNN_CONVOLUTION_FWD_ALGO_GEMM,
	CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD);


  // NOTE: Using GEMM-based Algo
  convAlgo = CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM;

  size_t workspace_size;
  checkCUDNN(cudnnGetConvolutionForwardWorkspaceSize(cudnnHandle,
						     input->tensor_desc,
						     filter->filter_desc,
						     convDesc,
						     output->tensor_desc,
						     convAlgo,
						     &workspace_size));

  // Allocating memory for the convolution workspace
  void* workspace;
  checkCudaErrors(cudaMalloc(&workspace, workspace_size));
  DEBUG("workspace size = %d \n", workspace_size);


  checkCUDNN(cudnnConvolutionForward(cudnnHandle, &alpha, input->tensor_desc,
				     input->gpu_data, filter->filter_desc, newFilter->gpu_data,
				     convDesc, convAlgo, workspace, workspace_size,
				     &beta, output->tensor_desc, output->gpu_data));


 
  freeTensor(newFilter);
  profileEvent("tensorConv_end", true);

  return output;
}









/************ NOTE: API for ApproxHPVM Wrapper runtime *******/ 


void* PROMISE_Conv(void* input, float i_min, float i_max,
		   void* filter, float w_min, float w_max,
		   void* bias, float b_min, float b_max,
		   int conv_pad_h, int conv_pad_w,
		   int conv_stride_h, int conv_stride_w,
		   int pool_id, int pool_size, int pool_stride,
		   int activation_id, // Relu, Tanh, ClipRelu
		   float out_min, float out_max, int swing){ 


  Tensor* input_t = (Tensor*) input;
  Tensor* filter_t = (Tensor*) filter;
  Tensor* bias_t = (Tensor*) bias;
  
  int orig_type = input_t->cur_type;

  DEBUG("FP32 conversions \n");
  
  convertToFP32(input_t);

  convertToFP32(filter_t);
  convertToFP32(bias_t);

  DEBUG("DONE FP32 conversions \n");
  

  if(swing < 8){
    input = quantizeTensorPromise(input, i_min, i_max);
    filter = quantizeTensorPromise(filter, w_min, w_max);
    if(bias != NULL)
      bias = quantizeTensorPromise(bias, b_min, b_max);
    // aRead error
    
    input = addPromiseError(input, swing);
  }

  
  void* conv_out;
  conv_out = tensorConvolution(input, filter,
			       conv_pad_h, conv_pad_w,
			       conv_stride_h, conv_stride_w,
			       1, 0);
  
  void* conv_add;
  if(bias != NULL){
    conv_add = tensorAdd(conv_out, bias);
  }
  else{
    conv_add = conv_out;
  }

  void* pool_out;
  // NOTE: Skip pooling on negative pool sizes
  if(pool_size > 0){
    //FIXME: Currently only using MaxPooling
    //-- pool_out = tensorPooling(conv_add, 0, pool_size, pool_size, 0, 0, pool_size, pool_size);
    pool_out = tensorPooling(conv_add, 0, pool_size, pool_size, 0, 0, pool_stride, pool_stride);
  }
  else{
    pool_out = conv_add;
  }
  
  void* activation_out;  
  switch(activation_id){
  case -1:
    activation_out = pool_out;
    INFO("NO Activation Function \n");
    break;
  case 0:
    activation_out = tensorTanh(pool_out);
    break;
  case 1:
    activation_out = tensorRelu(pool_out);
    break;
  case 2:
    activation_out = tensorRelu2(pool_out, out_min, out_max);
    break;
  default:
    ERROR("Activation id %d NOT supported \n", activation_out);
    break;
  }


  if(swing < 8 && activation_id != -1){
    activation_out = quantizeTensorPromise(activation_out, out_min, out_max);
  }



  //NOTE: Convert back to FP16 if original type
  if (orig_type == half_type){
    convertToFP16((Tensor*) activation_out);
  }

  
  return activation_out;
}



void* PROMISE_FC(void* input, float i_min, float i_max,
		 void* weights, float w_min, float w_max,
		 void* bias, float b_min, float b_max,
		 int activation_id,
		 float out_min, float out_max, int swing){


  Tensor* input_t = (Tensor*) input;
  Tensor* weights_t = (Tensor*) weights;
  Tensor* bias_t = (Tensor*) bias;
  
  int orig_type = input_t->cur_type;
  
  convertToFP32(input_t);
  convertToFP32(weights_t);
  convertToFP32(bias_t);
  
  
  if(swing < 8){
    input = quantizeTensorPromise(input, i_min, i_max);
    weights = quantizeTensorPromise(weights, w_min, w_max);
    if(bias != NULL)
      bias = quantizeTensorPromise(bias, b_min, b_max);

    // NOTE: Modelling aRead error in PROMISE
    input = addPromiseError(input, swing);
  }


  
  void* gemm_out;
  gemm_out = tensorGemmGPU(input, weights);

  
  void* gemmbias_out;
  if(bias != NULL){
    gemmbias_out = tensorAdd(gemm_out, bias);
  }
  else{
    gemmbias_out = gemm_out;
  }
 
  void* activation_out;
  switch(activation_id){

  case -1:
    activation_out = gemmbias_out;
    INFO("No Activation Function \n");
    break;
  case 0:
    activation_out = tensorTanh(gemmbias_out);
    break;
  case 1:
    activation_out = tensorRelu(gemmbias_out);
    break;
  case 2:
    activation_out = tensorRelu2(gemmbias_out, out_min, out_max);
    break;
  default:
    ERROR("Activation id %d NOT supported \n", activation_out);
    break;
  }
  
  
  if(swing < 8 && activation_id != -1){
    activation_out = quantizeTensorPromise(activation_out, out_min, out_max);
  }


  //NOTE: Convert back to FP16 if original type
  if (orig_type == half_type){
    convertToFP16((Tensor*) activation_out);
  }


  
  return activation_out;
}





// NOTE: Enabling the macro below is used for testing against the old PROMISE wrapper
//#define OLD_MODEL

#ifndef OLD_MODEL



bool isPromiseLayer(int swing){

  if(swing < 8)
    return true;
  else
    return false;      
}


bool isGPULayer(int swing){

  if(swing > 10 ) // PROMISE layers are 1-7
    return true;
  else
    return false;      
}


bool isFullPrecision(int swing){

  if(swing == 11)
    return true;
  else
    return false;      
}



bool isHalfPrecision(int swing){

  if(swing == 12)
    return true;
  else
    return false;      
}


bool isPerforation(int swing){

  if(swing >= 100 && swing <= 200)
    return true;
  else
    return false;      
}


bool isSampling(int swing){

  if(swing >= 200 && swing <= 300)
    return true;
  else
    return false;      
}

bool isReductionSampling(int swing) {

  if (swing >= 41 && swing <= 49)
    return true;
  else
    return false;      
}

int getSwing(int swing){

  #ifdef PROMISE_TUNER_ENABLED

  // NOTE: Skip reading file-based error levels for ApproxHPVM wrapper runtime
  if(!approxhpvm_runtime_mode){
  
    if(op_counter >= total_ops){
      ERROR("No accuracy flag found \n");
    }
  
    swing = op_accuracies[op_counter];
    op_counter++;
  }

  #endif  

   DEBUG("---- swing_value = %d \n", swing);  

   return swing;
}




//bool FP16_tuning = false;


/***** API for Autotuner Use - Not the ApproxHPVM Wrapper API */



void initializeAutotuner(){

  printf("initializing tuner .... \n");
  
  sampParamSet = new SampParamSet;
  perfParamSet = new PerfParamSet;  
}


void* Autotuner_SampConv(void* input, float i_min, float i_max,
			 void* filter, float w_min, float w_max,
			 void* bias, float b_min, float b_max,
			 int conv_pad_h, int conv_pad_w,
			 int conv_stride_h, int conv_stride_w,
			 int pool_id, int pool_size,
			 int activation_id, // Relu, Tanh, ClipRelu
			 float out_min, float out_max, int swing){


  SampParams params = sampParamSet->getSampParams(swing);
  
  DEBUG("params.skip_rate = %d, params.skip_offset = %d \n",
	params.skip_rate, params.skip_offset);
  
  void* conv_out;
  
  if (!FP16_tuning){
 
    /* conv_out = tensorConvSampSim(input, filter,
				 conv_pad_h, conv_pad_w,
				 conv_stride_h, conv_stride_w, 1, 1,
				 params.skip_rate, params.skip_offset);
    */


    if (SIMULATION_MODE){
      conv_out = tensorConvSampSim2(input, filter,
				    conv_pad_h, conv_pad_w,
				    conv_stride_h, conv_stride_w, 1, 1,
				    params.skip_rate, params.skip_offset, params.interpolation_id);
    }
        

    else {
      conv_out = tensorConvApprox(input, filter,
				  conv_pad_h, conv_pad_w,
				  conv_stride_h, conv_stride_w, 1, 1,
				  1, 1, params.skip_rate, params.skip_offset);
    }
    
    
  }
  else{
        
    conv_out = tensorConvApproxHalf2(input, filter,
				     conv_pad_h, conv_pad_w,
				     conv_stride_h, conv_stride_w,
				     1, 1,
				     1, 1,
				     params.skip_rate, params.skip_offset);    
   
  }

  return conv_out;
}




void* Autotuner_PerforatedConv(void* input, float i_min, float i_max,
			       void* filter, float w_min, float w_max,
			       void* bias, float b_min, float b_max,
			       int conv_pad_h, int conv_pad_w,
			       int conv_stride_h, int conv_stride_w,
			       int pool_id, int pool_size,
			       int activation_id, // Relu, Tanh, ClipRelu
			       float out_min, float out_max, int swing){ 


  PerfParams params = perfParamSet->getPerfParams(swing);
  
  DEBUG("params.row = %d, params.col = %d, params.skip_offset = %d \n",
	params.row, params.col, params.skip_offset);
    

  void* conv_out;
  
  if (!FP16_tuning){


    if (SIMULATION_MODE){

      conv_out = tensorConvPerfCuda(input, filter,
				    conv_pad_h, conv_pad_w,
				    conv_stride_h, conv_stride_w, 1, 1,
				    params.row, params.col, params.skip_offset);

    }
    else{

      conv_out = tensorConvApprox(input, filter,
				  conv_pad_h, conv_pad_w,
				  conv_stride_h, conv_stride_w,
				  1, 1,
				  params.row, params.col,
				  1, params.skip_offset);   
    }
    
    
  }
  else{

    conv_out = tensorConvApproxHalf2(input, filter,
				     conv_pad_h, conv_pad_w,
				     conv_stride_h, conv_stride_w,
				     1, 1,
				     params.row, params.col,
				     1, params.skip_offset);   

  }
    
  return conv_out;  
}






void* Autotuner_ConvOp(void* input, float i_min, float i_max,
		       void* filter, float w_min, float w_max,
		       void* bias, float b_min, float b_max,
		       int conv_pad_h, int conv_pad_w,
		       int conv_stride_h, int conv_stride_w,
		       int pool_id, int pool_size,
		       int activation_id, // Relu, Tanh, ClipRelu
		       float out_min, float out_max, int swing){ 

  
  void* conv_out;
  if(isPerforation(swing)){

    conv_out = Autotuner_PerforatedConv(input, i_min, i_max,
					filter, w_min, w_max,
					bias, b_min, b_max,
					conv_pad_h, conv_pad_w,
					conv_stride_h, conv_stride_w,
					pool_id, pool_size,
					activation_id, 
					out_min, out_max, swing);
    
  }

  else if(isSampling(swing)){

    conv_out = Autotuner_SampConv(input, i_min, i_max,
				  filter, w_min, w_max,
				  bias, b_min, b_max,
				  conv_pad_h, conv_pad_w,
				  conv_stride_h, conv_stride_w,
				  pool_id, pool_size,
				  activation_id, 
				  out_min, out_max, swing);
  }
  

  else if (isHalfPrecision(swing)){

    if (FP16_tuning){
 
      conv_out = tensorHalfConvolution(input, filter,
				     conv_pad_h, conv_pad_w,
				     conv_stride_h, conv_stride_w,
				     1, 0);
    }
    else{
      conv_out = tensorConvolution(input, filter,
				 conv_pad_h, conv_pad_w,
				 conv_stride_h, conv_stride_w,
				 1, 0);
    }
    
  }

  else if (isFullPrecision(swing)){
    conv_out = tensorConvolution(input, filter,
				 conv_pad_h, conv_pad_w,
				 conv_stride_h, conv_stride_w,
				 1, 0);
  }


  return conv_out;  
}



void* Autotuner_Add(void* input, void* bias, int swing){

  void* conv_add;
  if(bias != NULL){
    
    if( isFullPrecision(swing) || !(FP16_tuning) ){  
      conv_add = tensorAdd(input, bias);
    }
    else {
      conv_add = tensorHalfAdd(input, bias);
    }
  }
  else{
    conv_add = input;
  }

  return conv_add;
}



void* Autotuner_Pooling(void* input,
			int pool_size, int pool_stride,
			int swing){

  void* pool_out;
  
  if(pool_size > 0){
    
    //FIXME: Currently only using MaxPooling
    if( isFullPrecision(swing) || !(FP16_tuning) ){  
      pool_out = tensorPooling(input, 0, pool_size, pool_size,
			       0, 0, pool_stride, pool_stride);

    }
	
    else {
      pool_out = tensorHalfPooling(input, 0, pool_size, pool_size,
				   0, 0, pool_stride, pool_stride);  
    }
    
   
  }
  else{
    pool_out = input;
  }
  
  
  return pool_out;
}




void* Autotuner_Activation(void* input, int activation_id,
			   int out_min, int out_max, int swing){

  void* activation_out;

  if ( isFullPrecision(swing) || (!FP16_tuning) ){
    
    switch(activation_id){
    case -1:
      activation_out = input;
      INFO("NO Activation Function \n");
      break;
    case 0:
      activation_out = tensorTanh(input);
      break;
    case 1:
      activation_out = tensorRelu(input);
      break;
    case 2:
      activation_out = tensorRelu2(input, out_min, out_max);
      break;
    default:
      ERROR("Activation id %d NOT supported \n", activation_out);
      break;
    }
  }
   
  else{

    switch(activation_id){
    case -1:
      activation_out = input;
      INFO("NO Activation Function \n");
      break;
    case 0:
      activation_out = tensorHalfTanh(input);
      break;
    case 1:
      activation_out = tensorHalfRelu(input);
      break;
    case 2:
      activation_out = tensorHalfRelu2(input, out_min, out_max);
      break;
    default:
      ERROR("Activation id %d NOT supported \n", activation_out);
      break;
    }

  }


  return activation_out;
}


void* autotuner_tensorFft(void *input, bool inverse) {
  if(ONLINE_PROFILING){
    ERROR("Online Profiling cannot be enabled\n");
    abort();
  }

  int swing = 0;
  swing = getSwing(swing);

  if (isFullPrecision(swing)) {
    return tensorFft(input, inverse);
  }
  if (isHalfPrecision(swing)) {
    return tensorFftHalf(input, inverse);
  }

  ERROR("Unsupported autotuner flag for operation fft\n");
  abort();
  return NULL;
}


void* autotuner_tensorReduce(void *input, size_t axis, MathOp func) {
  if(ONLINE_PROFILING){
    ERROR("Online Profiling cannot be enabled\n");
    abort();
  }

  int swing = 0;
  swing = getSwing(swing);

  if (isFullPrecision(swing)) {
    return tensorReduce(input, axis, func, 0.0f);
  }

  if (isHalfPrecision(swing)) {
    return tensorReduceHalf(input, axis, func, 0.0f);
  }

  if (isReductionSampling(swing)) {
    RedSampParams params = getRedSampParams(swing);
    DEBUG("params.skip_ratio = %f, params.is_half = %d\n", params.skip_ratio, (int)params.is_half);
    if (params.is_half)
      return tensorReduceHalf(input, axis, func, params.skip_ratio);
    else
      return tensorReduce(input, axis, func, params.skip_ratio);
  }

  ERROR("Unsupported autotuner flag for operation reduce\n");
  abort();
  return NULL;
}

void* autotuner_tensorProjectiveT(void *input, void *transformation) {
  if(ONLINE_PROFILING){
    ERROR("Online Profiling cannot be enabled\n");
    abort();
  }

  int swing = 0;
  swing = getSwing(swing);

  if (isFullPrecision(swing)) {
    return tensorProjectiveT(input, transformation);
  }

  ERROR("Unsupported autotuner flag for operation projectiveT\n");
  abort();
  return NULL;
}


void* autotuner_tensorMap1(MathOp func, void *input) {
  if(ONLINE_PROFILING){
    ERROR("Online Profiling cannot be enabled\n");
    abort();
  }

  int swing = 0;
  swing = getSwing(swing);

  if (isFullPrecision(swing)) {
    return tensorMap1(func, input);
  }

  if (isHalfPrecision(swing)) {
    return tensorMap1Half(func, input);
  }

  ERROR("Unsupported autotuner flag for operation map1\n");
  abort();
  return NULL;
}

void* autotuner_tensorMap2(MathOp func, void *input1, void *input2) {
  if(ONLINE_PROFILING){
    ERROR("Online Profiling cannot be enabled\n");
    abort();
  }

  int swing = 0;
  swing = getSwing(swing);

  if (isFullPrecision(swing)) {
    return tensorMap2(func, input1, input2);
  }

  if (isHalfPrecision(swing)) {
    return tensorMap2Half(func, input1, input2);
  }

  ERROR("Unsupported autotuner flag for operation map2\n");
  abort();
  return NULL;
}

void* autotuner_tensorMap3(MathOp func, void *input1, void *input2,
                           void *input3) {
  if(ONLINE_PROFILING){
    ERROR("Online Profiling cannot be enabled\n");
    abort();
  }

  int swing = 0;
  swing = getSwing(swing);

  if (isFullPrecision(swing)) {
    return tensorMap3(func, input1, input2, input3);
  }

  if (isHalfPrecision(swing)) {
    return tensorMap3Half(func, input1, input2, input3);
  }

  ERROR("Unsupported autotuner flag for operation map3\n");
  abort();
  return NULL;
}




void* Autotuner_GPU_ConvLayer(void* input, float i_min, float i_max,
			      void* filter, float w_min, float w_max,
			      void* bias, float b_min, float b_max,
			      int conv_pad_h, int conv_pad_w,
			      int conv_stride_h, int conv_stride_w,
			      int pool_id, int pool_size, int pool_stride, 
			      int activation_id, // Relu, Tanh, ClipRelu
			      float out_min, float out_max, int swing){ 
  

  void* conv_out = Autotuner_ConvOp(input, i_min, i_max,
				    filter, w_min, w_max,
				    bias, b_min, b_max,
				    conv_pad_h, conv_pad_w,
				    conv_stride_h, conv_stride_w,
				    pool_id, pool_size,
				    activation_id, 
				    out_min, out_max, swing);
 
  
  void* conv_add = Autotuner_Add(conv_out, bias, swing);

  void* pool_out = Autotuner_Pooling(conv_add, pool_size, pool_stride, swing);

  void* activation_out = Autotuner_Activation(pool_out, activation_id, out_min, out_max, swing);
  

  return activation_out;  
}


/**** Top-level API for Handling Convolution Layers

      The granularity of handling is at a layer-level - not tensor-op level
        
***/

void* Autotuner_ConvLayer(void* input, float i_min, float i_max,
			  void* filter, float w_min, float w_max,
			  void* bias, float b_min, float b_max,
			  int conv_pad_h, int conv_pad_w,
			  int conv_stride_h, int conv_stride_w,
			  int pool_id, int pool_size, int pool_stride, 
			  int activation_id, // Relu, Tanh, ClipRelu
			  float out_min, float out_max, int swing){ 

  if(FP16_tuning){
    if(ONLINE_PROFILING){
      ERROR("Online Profiling cannot be enabled with PROMISE Simulation \n");
    }
  }

  swing = getSwing(swing);  
 
  if(isPromiseLayer(swing)){
    
    return PROMISE_Conv(input, i_min, i_max,
			filter, w_min, w_max,
			bias, b_min, b_max,
			conv_pad_h, conv_pad_w,
			conv_stride_h, conv_stride_w,
			pool_id, pool_size, pool_stride,
			activation_id, 
			out_min, out_max, swing);
  }

  assert(isGPULayer(swing));

  return Autotuner_GPU_ConvLayer(input, i_min, i_max,
			   filter, w_min, w_max,
			   bias, b_min, b_max,
			   conv_pad_h, conv_pad_w,
			   conv_stride_h, conv_stride_w,
			   pool_id, pool_size, pool_stride, 
			   activation_id,
			   out_min, out_max, swing);

}





/**** Top-level API Unchanged for backwards compatibility  ***/

void* ConvLayer_PROMISE(void* input, float i_min, float i_max,
			void* filter, float w_min, float w_max,
			void* bias, float b_min, float b_max,
			int conv_pad_h, int conv_pad_w,
			int conv_stride_h, int conv_stride_w,
			int pool_id, int pool_size,
			int activation_id, // Relu, Tanh, ClipRelu
			float out_min, float out_max, int swing){ 


  return Autotuner_ConvLayer(input, i_min, i_max,
			     filter, w_min, w_max,
			     bias, b_min, b_max,
			     conv_pad_h, conv_pad_w,
			     conv_stride_h, conv_stride_w,
			     pool_id, pool_size, pool_size, // FIXIT: Assumption pool_size == pool_strides
			     activation_id,
			     out_min, out_max, swing);
  

}




void* ConvLayer_PROMISE2(void* input, float i_min, float i_max,
			 void* filter, float w_min, float w_max,
			 void* bias, float b_min, float b_max,
			 int conv_pad_h, int conv_pad_w,
			 int conv_stride_h, int conv_stride_w,
			 int pool_id, int pool_size, int pool_stride, 
			 int activation_id, // Relu, Tanh, ClipRelu
			 float out_min, float out_max, int swing){ 


  return Autotuner_ConvLayer(input, i_min, i_max,
			     filter, w_min, w_max,
			     bias, b_min, b_max,
			     conv_pad_h, conv_pad_w,
			     conv_stride_h, conv_stride_w,
			     pool_id, pool_size, pool_stride, 
			     activation_id,
			     out_min, out_max, swing);
  

}







void* FCLayer_PROMISE(void* input, float i_min, float i_max,
		      void* weights, float w_min, float w_max,
		      void* bias, float b_min, float b_max,
		      int activation_id,
		      float out_min, float out_max, int swing){ //NOTE: min_val, max_val apply to 'ClippedRelu'


  swing = getSwing(swing);
  
  if(isPromiseLayer(swing)){

    return PROMISE_FC(input, i_min, i_max,
		      weights, w_min, w_max,
		      bias, b_min, b_max,
		      activation_id,
		      out_min, out_max, swing);
  }

  assert(isGPULayer(swing));

  void *gemm_out;
  if (!isFullPrecision(swing)) {
    gemm_out = tensorHalfGemm(input, weights);
  } else {
    gemm_out = tensorGemmGPU(input, weights);
  }

  void *gemmbias_out;
  if (bias != NULL) {
    // Swing 8 corresponds to FP32
    if (isFullPrecision(swing) || (!FP16_tuning)) {
      gemmbias_out = tensorAdd(gemm_out, bias);
    } else {
      gemmbias_out = tensorHalfAdd(gemm_out, bias);
    }
  } else {
    gemmbias_out = gemm_out;
  }

  void *activation_out;
  switch (activation_id) {

  case -1:
    activation_out = gemmbias_out;
    INFO("No Activation Function \n");
    break;
  case 0:
    activation_out = tensorTanh(gemmbias_out);
    break;
  case 1:
    activation_out = tensorRelu(gemmbias_out);
    break;
  case 2:
    activation_out = tensorRelu2(gemmbias_out, out_min, out_max);
    break;
  default:
    ERROR("Activation id %d NOT supported \n", activation_out);
    break;
  }

  return activation_out;

}

#endif



#ifdef OLD_MODEL

#endif

#endif 



