/* device_math.cu defines */
#include "device_math.h"
#include "error.h"

#include <thrust/complex.h>

#define DEF_FUNC_PTR(fname) __device__ void *fname##_ptr = (void *)(fname);

#define DEF_FUNC_PTR_CAST(fname, castf)                                        \
  __device__ void *fname##_ptr = (void *)(castf);

#define CASE_FUNC(ename, fname)                                                \
  case MathOp::ename: {                                                        \
    void *v_func_ptr = nullptr;                                                \
    checkCudaErrors(cudaMemcpyFromSymbol(                                      \
        &v_func_ptr, _internal::fname##_ptr, sizeof(void *)));                 \
    return v_func_ptr;                                                         \
  }

namespace _internal {

// The following functions are not used, but they reference their cuda
// counterpart which is necessary, otherwise ptx compilation breaks
__device__ float hypotf_(float x, float y) { return hypotf(x, y); }

__device__ float atan2f_(float x, float y) { return atan2f(x, y); }

__device__ float sqrtf_(float x) { return sqrtf(x); }

__device__ float fmax_(float x, float y) { return fmax(x, y); }

__device__ float fmin_(float x, float y) { return fmin(x, y); }

__device__ float add(float x, float y) { return x + y; }

__device__ float sub(float x, float y) { return x - y; }

__device__ float mul(float x, float y) { return x * y; }

__device__ float div(float x, float y) { return x / y; }

__device__ float favg3(float x) { return __fdividef(x, 3.0f); }

__device__ float blend2(float bg, float fg) { return bg * 0.6 + fg * 0.4; }

__device__ float addWeighted(float blurred, float image) {
  return 0.7 * image + 0.3 * blurred;
}

__device__ float psnr(float x) { return -10 * log10(x); }

__device__ float2 f2mul(float2 x1, float2 x2) {
  return {x1.x * x2.x - x1.y * x2.y, x1.x * x2.y + x1.y * x2.x};
}

__device__ half2 h2mul(half2 x1, half2 x2) {
  return {x1.x * x2.x - x1.y * x2.y, x1.x * x2.y + x1.y * x2.x};
}

__device__ half2 h2hypot(half2 x, half2 y) {
  return h2sqrt(__hfma2(x, x, __hmul2(y, y)));
}

__device__ half2 h2max(half2 x, half2 y) {
  return __hfma2(__hgt2(x, y), x, __hmul2(__hle2(x, y), y));
}

__device__ half2 h2min(half2 x, half2 y) {
  return __hfma2(__hlt2(x, y), x, __hmul2(__hge2(x, y), y));
}

__device__ half2 h2avg3(half2 x) {
  half2 three = __floats2half2_rn(3.0f, 3.0f);
  return __h2div(x, three);
}

__device__ half2 h2blend2(half2 bg, half2 fg) {
  half2 c1 = __floats2half2_rn(0.6f, 0.6f), c2 = __floats2half2_rn(0.4f, 0.4f);
  return __hfma2(bg, c1, __hmul2(fg, c2));
}

__device__ half2 h2addWeighted(half2 blurred, half2 image) {
  half2 c1 = __floats2half2_rn(0.7f, 0.7f), c2 = __floats2half2_rn(0.3f, 0.3f);
  return __hfma2(image, c1, __hmul2(blurred, c2));
}

DEF_FUNC_PTR(hypotf)
DEF_FUNC_PTR(atan2f)
DEF_FUNC_PTR(add)
DEF_FUNC_PTR(sub)
DEF_FUNC_PTR(mul)
DEF_FUNC_PTR(div)
DEF_FUNC_PTR(sqrtf)
DEF_FUNC_PTR_CAST(fmax, (float (*)(float, float))fmax)
DEF_FUNC_PTR_CAST(fmin, (float (*)(float, float))fmin)
DEF_FUNC_PTR(favg3)
DEF_FUNC_PTR(blend2)
DEF_FUNC_PTR(addWeighted)
DEF_FUNC_PTR(psnr)

DEF_FUNC_PTR(f2mul)

DEF_FUNC_PTR(h2mul)

DEF_FUNC_PTR(h2hypot)
DEF_FUNC_PTR(__hadd2)
DEF_FUNC_PTR(__hsub2)
DEF_FUNC_PTR(__h2div)
DEF_FUNC_PTR(h2sqrt)
DEF_FUNC_PTR(h2max)
DEF_FUNC_PTR(h2min)
DEF_FUNC_PTR(h2avg3)
DEF_FUNC_PTR(h2blend2)
DEF_FUNC_PTR(h2addWeighted)

} // namespace _internal

template <> void *mathOpToFunc<float2>(MathOp op) {
  switch (op) {
    CASE_FUNC(Mul, f2mul)
  default:
    ERROR("Float2 function not found\n");
    return nullptr;  // For some compilers
  }
}

template <> void *mathOpToFunc<half2>(MathOp op) {
  switch (op) {
    CASE_FUNC(Mul, h2mul)
  default:
    ERROR("Half2 function not found\n");
    return nullptr;  // For some compilers
  }
}

template <> void *mathOpToFunc<float>(MathOp op) {
  switch (op) {
    CASE_FUNC(Hypot, hypotf)
    CASE_FUNC(Atan2, atan2f)
    CASE_FUNC(Add, add)
    CASE_FUNC(Sub, sub)
    CASE_FUNC(Mul, mul)
    CASE_FUNC(Div, div)
    CASE_FUNC(Sqrt, sqrtf)
    CASE_FUNC(Max, fmax)
    CASE_FUNC(Min, fmin)
    CASE_FUNC(Avg3, favg3)
    CASE_FUNC(Blend2, blend2)
    CASE_FUNC(AddWeighted, addWeighted)
    CASE_FUNC(PSNR, psnr)
  default:
    ERROR("Float function not found\n");
  }
  return nullptr;  // For some compilers
}

template <> void *mathOpToFunc<half>(MathOp op) {
  switch (op) {
    CASE_FUNC(Hypot, h2hypot)
    CASE_FUNC(Add, __hadd2)
    CASE_FUNC(Sub, __hsub2)
    CASE_FUNC(Div, __h2div)
    CASE_FUNC(Sqrt, h2sqrt)
    CASE_FUNC(Max, h2max)
    CASE_FUNC(Min, h2min)
    CASE_FUNC(Avg3, h2avg3)
    CASE_FUNC(Blend2, h2blend2)
    CASE_FUNC(AddWeighted, h2addWeighted)
  default:
    ERROR("Half function not found\n");
  }
  return nullptr;  // For some compilers
}

template <> half reduceOpToIdentity<half>(MathOp op) {
  switch (op) {
  case MathOp::Hypot:
    return 0.0f;
  case MathOp::Add:
    return 0.0f;
  case MathOp::Max:
    return -65504.0f;
  case MathOp::Min:
    return 65504.0f;
  default:
    ERROR("Operator does not have id value\n");
  }
  return 0.0f;  // For some compilers
}
