#include "approxhpvm_img_runtime_utils.h"
#include "debug.h"
#include "img_tensor_runtime.h"

#include "functional/map.cuh"
#include "functional/reduce.cuh"
#include "tensor_utils.h"

#include <cufft.h>
#include <cufftXt.h>
#include <thrust/device_vector.h>

template <typename T> struct DivFunctor {
  const T dividend;

  DivFunctor(float dividend) : dividend(dividend) {}

  __host__ __device__ T operator()(const T x) const { return x / dividend; }
};

// ***                       Runtime implementation                      *** //
void *tensorFft(void *input, bool inverse) {
  // https://docs.nvidia.com/cuda/cufft/index.html#twod-complex-to-real-transforms
  // Tensor checking
  INFO("FFT\n");
  profileEvent("tensorFft");
  auto *t_input = (Tensor *)input;
  int total_rank = t_input->dims.num_dims;
  if (total_rank != 4)
    ERROR("Only 4-dim tensor supported\n");
  // Dimensions
  size_t *all_dims = t_input->dims.dim_sizes;
  int height = all_dims[2], width = all_dims[3];
  int n_batch = int(all_dims[0]) * int(all_dims[1]);
  // Prepare input data
  hostToDeviceCopy(t_input);
  // Create a 2D FFT plan
  cufftHandle plan;
  checkCUFFT(cufftCreate(&plan));
  // Output
  Tensor *out_tensor = nullptr;
  if (inverse) {
    int fft_dim[2] = {height, (width - 1) * 2};
    auto *input_cuda = convertAndGetGPUData<cufftComplex>(t_input);
    // Define output data
    out_tensor = (Tensor *)create4DTensor(
        (int)float_type, CUDNN_TENSOR_NCHW, all_dims[0], all_dims[1], height,
        (width - 1) * 2);
    changeTensorPlacement(out_tensor, DEVICE);
    auto *output_cuda = convertAndGetGPUData<cufftReal>(out_tensor);
    checkCUFFT(cufftPlanMany(
        &plan, 2, fft_dim, nullptr, 1, 0, nullptr, 1, 0, CUFFT_C2R, n_batch));
    // Execute the plan
    checkCUFFT(cufftExecC2R(plan, input_cuda, output_cuda));
  } else {
    int fft_dim[2] = {height, width};
    auto *input_cuda = convertAndGetGPUData<cufftReal>(t_input);
    // Define output data
    out_tensor = (Tensor *)create4DTensor(
        (int)float2_type, CUDNN_TENSOR_NCHW, all_dims[0], all_dims[1], height,
        (width / 2 + 1));
    changeTensorPlacement(out_tensor, DEVICE);
    auto *output_cuda = convertAndGetGPUData<cufftComplex>(out_tensor);
    checkCUFFT(cufftPlanMany(
        &plan, 2, fft_dim, nullptr, 1, 0, nullptr, 1, 0, CUFFT_R2C, n_batch));
    // Execute the plan
    checkCUFFT(cufftExecR2C(plan, input_cuda, output_cuda));
  }
  // Wait for the device to finish
  checkCUDA(cudaDeviceSynchronize());

  if (inverse) {
    auto *output_cuda = convertAndGetGPUData<cufftReal>(out_tensor);
    thrust::device_ptr<cufftReal> normalize_v(output_cuda);
    size_t size = height * (width - 1) * 2;
    DivFunctor<cufftReal> div(size);
    thrust::transform(normalize_v, normalize_v + size, normalize_v, div);
  }
  // Release memory
  cufftDestroy(plan);
  profileEvent("tensorFft_end");
  return out_tensor;
}

void *tensorFftHalf(void *input, bool inverse) {
  // Tensor checking
  INFO("FFTHalf\n");
  profileEvent("#tensorFft");
  auto *t_input = (Tensor *)input;
  int total_rank = t_input->dims.num_dims;
  if (total_rank != 4)
    ERROR("Only 4-dim tensor supported\n");
  // Dimensions
  size_t *all_dims = t_input->dims.dim_sizes;
  int height = all_dims[2], width = all_dims[3];
  long long int n_batch = int(all_dims[0]) * int(all_dims[1]);
  // Prepare input data
  hostToDeviceCopy(t_input);
  // Create a 2D FFT plan
  cufftHandle plan;
  checkCUFFT(cufftCreate(&plan));
  // Output
  Tensor *out_tensor = nullptr;
  if (inverse) {
    long long int fft_dim[2] = {height, (width - 1) * 2};
    profileEvent("F2H_start");
    auto *input_cuda = convertAndGetGPUData<half2>(t_input);
    profileEvent("F2H_end");
    // Define output data
    out_tensor = (Tensor *)create4DTensor(
        (int)half_type, CUDNN_TENSOR_NCHW, all_dims[0], all_dims[1], height,
        (width - 1) * 2);
    changeTensorPlacement(out_tensor, DEVICE);
    auto *output_cuda = convertAndGetGPUData<half>(out_tensor);
    size_t worksize = 0;
    checkCUFFT(cufftXtMakePlanMany(
        plan, 2, fft_dim, nullptr, 1, 0, CUDA_C_16F /*inputtype*/, nullptr, 1,
        0, CUDA_R_16F /*outputtype*/, n_batch, &worksize,
        CUDA_C_16F /*executiontype*/
        ));
    // Execute the plan
    checkCUFFT(cufftXtExec(plan, input_cuda, output_cuda, CUFFT_INVERSE));
  } else {
    long long int fft_dim[2] = {height, width};
    profileEvent("F2H_start");
    auto *input_cuda = convertAndGetGPUData<half>(t_input);
    profileEvent("F2H_end");
    // Define output data
    out_tensor = (Tensor *)create4DTensor(
        (int)half2_type, CUDNN_TENSOR_NCHW, all_dims[0], all_dims[1], height,
        (width / 2 + 1));
    changeTensorPlacement(out_tensor, DEVICE);
    auto *output_cuda = convertAndGetGPUData<half2>(out_tensor);
    size_t worksize = 0;
    checkCUFFT(cufftXtMakePlanMany(
        plan, 2, fft_dim, nullptr, 1, 0, CUDA_R_16F /*inputtype*/, nullptr, 1,
        0, CUDA_C_16F /*outputtype*/, n_batch, &worksize,
        CUDA_C_16F /*executiontype*/
        ));
    // Execute the plan
    checkCUFFT(cufftXtExec(plan, input_cuda, output_cuda, CUFFT_FORWARD));
  }
  // Wait for the device to finish
  checkCUDA(cudaDeviceSynchronize());

  if (inverse) {
    auto *output_cuda = convertAndGetGPUData<half>(out_tensor);
    thrust::device_ptr<half> normalize_v(output_cuda);
    size_t size = height * (width - 1) * 2;
    DivFunctor<half> div(size);
    thrust::transform(normalize_v, normalize_v + size, normalize_v, div);

    profileEvent("H2F_start");
    convertToFP32_offline(out_tensor);
    out_tensor->data_type = out_tensor->cur_type;
    profileEvent("H2F_end");
  } else {
    profileEvent("H2F_start");
    convertToFloat2Offline(out_tensor);
    profileEvent("H2F_end");
  }
  // Release memory
  cufftDestroy(plan);

  profileEvent("#tensorFft_end");
  return out_tensor;
}

/* Implements template instantiations in reduce.cuh */

template <>
__device__ void reduceAlongDim<float>(
    float *target, float *src, float init, float identity, void *func,
    size_t num_irows, size_t dim_size) {
  auto *binary_op = (NTo1MapF<float, 2>)func;

  float acc = init;
  for (size_t col = 0; col < dim_size; ++col) {
    acc = binary_op(acc, *src);
    src += num_irows;
  }
  *target = acc;
}

template <>
__device__ void parallelReduceAlongDim<float>(
    float *target, float *src, float init, float identity, void *func,
    size_t num_irows, size_t dim_size, size_t along_dim_tid,
    size_t n_along_dim_threads) {
  __shared__ float sbuf[CrossDimTh][AlongDimTh + 1]; // avoid bank conflict
  float *this_buf_line = sbuf[threadIdx.y];

  auto *binary_op = (NTo1MapF<float, 2>)func;

  float acc = init;
  // Sequential reduction within a thread.
  for (size_t col = along_dim_tid; col < dim_size; col += n_along_dim_threads) {
    acc = binary_op(acc, src[col * num_irows]);
  }

  this_buf_line[along_dim_tid] = acc;

  __syncthreads();

  // Reduce intermediate values to single value.
  for (size_t s = AlongDimTh >> 1u; s > 0; s >>= 1u) {
    if (along_dim_tid < s) {
      float arg1 = this_buf_line[along_dim_tid];
      float arg2 = this_buf_line[along_dim_tid + s];
      float res = binary_op(arg1, arg2);
      this_buf_line[along_dim_tid] = res;
    }
    __syncthreads();
  }

  if (along_dim_tid == 0) {
    *target = this_buf_line[0];
  }
  __syncthreads();
}

static __device__ __forceinline__ half
reduceHalf2ToHalf(NTo1MapF<half2, 2> half2_f, half2 value) {
  half2 high = __high2half2(value), low = __low2half2(value);
  return __high2half(half2_f(high, low)); // Or the lower half, whatever
}

template <>
__device__ void reduceAlongDim<half>(
    half *target, half *src, half init, half identity, void *func,
    size_t num_irows, size_t dim_size) {
  auto *binary_op = (NTo1MapF<half2, 2>)func;

  half2 acc = __halves2half2(init, identity);
  size_t twice_irows = num_irows << 1;
  for (size_t col = 0; col < dim_size; col += 2) {
    half higher = col + 1 < dim_size ? *(src + num_irows) : identity;
    acc = binary_op(acc, __halves2half2(*src, higher));
    src += twice_irows;
  }
  *target = reduceHalf2ToHalf(binary_op, acc);
}

template <>
__device__ void parallelReduceAlongDim<half>(
    half *target, half *src, half init, half identity, void *func,
    size_t num_irows, size_t dim_size, size_t along_dim_tid,
    size_t n_along_dim_threads) {
  __shared__ half2 sbuf[CrossDimTh][AlongDimTh + 1]; // avoid bank conflict
  half2 *this_buf_line = sbuf[threadIdx.y];

  auto *binary_op = (NTo1MapF<half2, 2>)func;

  // Sequential reduction within a thread.
  half2 acc = __halves2half2(init, identity);
  size_t src_stride = n_along_dim_threads * num_irows;
  for (size_t col = along_dim_tid; col < dim_size;
       col += (n_along_dim_threads << 1), src += src_stride << 1) {
    half higher =
        col + n_along_dim_threads < dim_size ? *(src + src_stride) : identity;
    acc = binary_op(acc, __halves2half2(*src, higher));
  }

  this_buf_line[along_dim_tid] = acc;

  __syncthreads();

  // Reduce intermediate values to single value.
  for (size_t s = AlongDimTh >> 1u; s > 0; s >>= 1u) {
    if (along_dim_tid < s) {
      half2 arg1 = this_buf_line[along_dim_tid];
      half2 arg2 = this_buf_line[along_dim_tid + s];
      half2 res = binary_op(arg1, arg2);
      this_buf_line[along_dim_tid] = res;
    }
    __syncthreads();
  }

  if (along_dim_tid == 0) {
    *target = reduceHalf2ToHalf(binary_op, this_buf_line[0]);
  }
  __syncthreads();
}

template <>
__global__ void kernelMapBroadcast<float, 1>(
    float *target, unsigned num_rows, void *func, float **srcs,
    size_t *tail_strides) {
  auto *n_ary_op = (NTo1MapF<float, 1>)func;

  unsigned threadId = blockIdx.x * blockDim.x + threadIdx.x,
           stride = gridDim.x * blockDim.x;
  for (unsigned row = threadId; row < num_rows; row += stride) {
    target[row] = n_ary_op(srcs[0][row]);
  }
}

template <>
__global__ void kernelMapBroadcast<float, 2>(
    float *target, unsigned num_rows, void *func, float **srcs,
    size_t *tail_strides) {
  auto *n_ary_op = (NTo1MapF<float, 2>)func;

  unsigned threadId = blockIdx.x * blockDim.x + threadIdx.x,
           stride = gridDim.x * blockDim.x;
  for (unsigned row = threadId; row < num_rows; row += stride) {
    unsigned j0 = row / tail_strides[0], j1 = row / tail_strides[1];
    target[row] = n_ary_op(srcs[0][j0], srcs[1][j1]);
  }
}

template <>
__global__ void kernelMapBroadcast<float, 3>(
    float *target, unsigned num_rows, void *func, float **srcs,
    size_t *tail_strides) {
  auto *n_ary_op = (NTo1MapF<float, 3>)func;

  unsigned threadId = blockIdx.x * blockDim.x + threadIdx.x,
           stride = gridDim.x * blockDim.x;
  for (unsigned row = threadId; row < num_rows; row += stride) {
    unsigned j0 = row / tail_strides[0], j1 = row / tail_strides[1],
             j2 = row / tail_strides[2];
    target[row] = n_ary_op(srcs[0][j0], srcs[1][j1], srcs[1][j2]);
  }
}

template <>
__global__ void kernelMapBroadcast<half, 1>(
    half *target, unsigned num_rows, void *func, half **srcs,
    size_t *tail_strides) {
  auto *op = (NTo1MapF<half2, 1>)func;

  unsigned threadId = (blockIdx.x * blockDim.x + threadIdx.x) << 1,
           stride = (gridDim.x * blockDim.x) << 1;
  unsigned row = threadId;
  for (; row < num_rows - 1; row += stride) {
    auto *m0 = (half2 *)&srcs[0][row], *m_out = (half2 *)&target[row];
    *m_out = op(*m0);
  }
  if (row == num_rows - 1) {
    half2 result = op(__half2half2(srcs[0][row]));
    target[row] = __high2half(result);
  }
}

template <>
__global__ void kernelMapBroadcast<half, 2>(
    half *target, unsigned num_rows, void *func, half **srcs,
    size_t *tail_strides) {
  auto *op = (NTo1MapF<half2, 2>)func;

  unsigned o_row = (blockIdx.x * blockDim.x + threadIdx.x) << 1,
           o_stride = gridDim.x * blockDim.x, o_stride2 = o_stride << 1;
  for (; o_row < num_rows - 1; o_row += o_stride2) {
    half2 *o_ptr = (half2 *)&target[o_row];
    half2 in[2];
    for (size_t i = 0; i < 2; i++) {
      if (tail_strides[i] == 1) {
        in[i] = __halves2half2(srcs[i][o_row], srcs[i][o_row + 1]);
      } else {
        unsigned i_row_l = o_row / tail_strides[i],
                 i_row_r = (o_row + 1) / tail_strides[i];
        if (i_row_l == i_row_r)
          in[i] = __half2half2(srcs[i][i_row_l]);
        else
          in[i] = __halves2half2(srcs[i][i_row_l], srcs[i][i_row_r]);
      }
    }
    *o_ptr = op(in[0], in[1]);
  }
  if (o_row == num_rows - 1) {
    unsigned row0 = o_row / tail_strides[0], row1 = o_row / tail_strides[1];
    half2 v0 = __half2half2(srcs[0][row0]), v1 = __half2half2(srcs[1][row1]);
    half2 result = op(v0, v1);
    target[o_row] = __high2half(result);
  }
}

template <>
__global__ void kernelMapBroadcast<half, 3>(
    half *target, unsigned num_rows, void *func, half **srcs,
    size_t *tail_strides) {
  auto *op = (NTo1MapF<half2, 3>)func;

  unsigned threadId = (blockIdx.x * blockDim.x + threadIdx.x) << 1,
           stride = (gridDim.x * blockDim.x) << 1;
  unsigned row = threadId;
  for (; row < num_rows - 1; row += stride) {
    unsigned row0 = row / tail_strides[0], row1 = row / tail_strides[1],
             row2 = row / tail_strides[2];
    auto *m0 = (half2 *)&srcs[0][row0], *m1 = (half2 *)&srcs[1][row1],
         *m2 = (half2 *)&srcs[2][row2], *m_out = (half2 *)&target[row];
    *m_out = op(*m0, *m1, *m2);
  }
  if (row == num_rows - 1) {
    unsigned row0 = row / tail_strides[0], row1 = row / tail_strides[1],
             row2 = row / tail_strides[2];
    half2 v0 = __half2half2(srcs[0][row0]), v1 = __half2half2(srcs[1][row1]),
          v2 = __half2half2(srcs[2][row2]);
    half2 result = op(v0, v1, v2);
    target[row] = __high2half(result);
  }
}

void *tensorReduce(void *input, size_t axis, MathOp func, float skip_ratio) {
  INFO("Reduce\n");
  profileEvent("tensorReduce");
  auto *src = (Tensor *)input;
  if (axis >= src->dims.num_dims)
    ERROR("Dimension out of range\n");
  if (src->dims.num_dims != 4 || src->data_format != CUDNN_TENSOR_NCHW)
    ERROR("Not supported\n");
  Tensor *ret = reduceDim<float>(src, 0.0f, func, axis, skip_ratio);
  profileEvent("tensorReduce_end");
  return ret;
}

void *
tensorReduceHalf(void *input, size_t axis, MathOp func, float skip_ratio) {
  INFO("Reduce\n");
  profileEvent("#tensorReduce");
  auto *src = (Tensor *)input;
  if (axis >= src->dims.num_dims)
    ERROR("Dimension out of range\n");
  if (src->dims.num_dims != 4 || src->data_format != CUDNN_TENSOR_NCHW)
    ERROR("Not supported\n");
  Tensor *ret = reduceDim<half>(src, 0.0f, func, axis, skip_ratio);
  profileEvent("H2F_start");
  convertToFP32_offline(ret);
  profileEvent("H2F_end");
  profileEvent("#tensorReduce_end");
  return ret;
}

void *tensorProjectiveT(void *input, void *transformation) {
  ERROR("ProjectiveT operation currently unsupported.\n");
  abort();
}

void *tensorMap1(MathOp f, void *i) {
  INFO("Map1\n");
  profileEvent("tensorMap1");
  auto *src = (Tensor *)i;
  Tensor *ret = mapGeneral<float, 1>(f, {src});
  profileEvent("tensorMap1_end");
  return ret;
}

void *tensorMap1Half(MathOp f, void *i) {
  INFO("Map1Half\n");
  profileEvent("#tensorMap1");
  auto *src = (Tensor *)i;
  Tensor *ret = mapGeneral<half, 1>(f, {src});
  profileEvent("H2F_start");
  convertToFP32_offline(ret);
  profileEvent("H2F_end");
  profileEvent("#tensorMap1_end");
  return ret;
}

void *tensorMap2(MathOp f2, void *i1, void *i2) {
  INFO("Map2\n");
  profileEvent("tensorMap2");
  auto *src1 = (Tensor *)i1, *src2 = (Tensor *)i2;
  Tensor_type_t common_ty =
      getCompatibleType(src1->cur_type, src2->cur_type, false);
  Tensor *ret = nullptr;
  if (common_ty == float_type)
    ret = mapGeneral<float, 2>(f2, {src1, src2});
  else if (common_ty == float2_type)
    ret = mapGeneral<float2, 2>(f2, {src1, src2});
  else
    ERROR("Type not recognized\n");
  profileEvent("tensorMap2_end");
  return ret;
}

void *tensorMap2Half(MathOp f2, void *i1, void *i2) {
  INFO("Map2Half\n");
  profileEvent("#tensorMap2");
  auto *src1 = (Tensor *)i1, *src2 = (Tensor *)i2;
  Tensor_type_t common_ty =
      getCompatibleType(src1->cur_type, src2->cur_type, false);
  if (common_ty == float_type) {
    Tensor *ret = mapGeneral<half, 2>(f2, {src1, src2});
    profileEvent("H2F_start");
    convertToFP32_offline(ret);
    profileEvent("H2F_end");
    profileEvent("#tensorMap2_end");
    return ret;
  } else if (common_ty == float2_type) {
    Tensor *ret = mapGeneral<half2, 2>(f2, {src1, src2});
    profileEvent("H2F_start");
    convertToFloat2Offline(ret);
    profileEvent("H2F_end");
    profileEvent("#tensorMap2_end");
    return ret;
  } else {
    ERROR("Type not recognized\n");
    return nullptr; // For some compilers
  }
}

void *tensorMap3(MathOp f3, void *i1, void *i2, void *i3) {
  INFO("Map3\n");
  profileEvent("tensorMap3");
  auto *src1 = (Tensor *)i1, *src2 = (Tensor *)i2, *src3 = (Tensor *)i3;
  Tensor *ret = mapGeneral<float, 3>(f3, {src1, src2, src3});
  profileEvent("tensorMap3_end");
  return ret;
}

void *tensorMap3Half(MathOp f3, void *i1, void *i2, void *i3) {
  INFO("Map3Half\n");
  profileEvent("#tensorMap3");
  auto *src1 = (Tensor *)i1, *src2 = (Tensor *)i2, *src3 = (Tensor *)i3;
  Tensor *ret = mapGeneral<half, 3>(f3, {src1, src2, src3});
  profileEvent("H2F_start");
  convertToFP32_offline(ret);
  profileEvent("H2F_end");
  profileEvent("#tensorMap3_end");
  return ret;
}

// ***                     Wrapper API implementation                    *** //

void *wrapper_tensorFft(const char *hpvm_node_id, void *input, bool inverse) {
  GPUNodeConfiguration *GPUConf =
      (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);
  std::vector<std::pair<
      GPUNodeConfiguration::TENSOR_OP,
      std::vector<std::pair<GPUNodeConfiguration::APPROX, int>>>>
      &ApproxChoices = GPUConf->getApproxChoices();
  // Approximation choices must be for a fft operation
  CUSTOM_ASSERT(
      ApproxChoices.size() == 1 &&
      ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::FFT &&
      "Invalid configuration generated for tensor fft wrapper operation");
  return handleTensorFftApproximationTuples(
      ApproxChoices[0].second, input, inverse);
}

void *wrapper_tensorReduce(
    const char *hpvm_node_id, void *input, int axis, int func) {
  GPUNodeConfiguration *GPUConf =
      (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);
  std::vector<std::pair<
      GPUNodeConfiguration::TENSOR_OP,
      std::vector<std::pair<GPUNodeConfiguration::APPROX, int>>>>
      &ApproxChoices = GPUConf->getApproxChoices();
  // Approximation choices must be for a reduce operation
  CUSTOM_ASSERT(
      ApproxChoices.size() == 1 &&
      ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::REDUCE &&
      "Invalid configuration generated for tensor reduce wrapper operation");
  return handleTensorReduceApproximationTuples(
      ApproxChoices[0].second, input, axis, (MathOp)func);
}

void *wrapper_tensorProjectiveT(
    const char *hpvm_node_id, void *input, void *transformation) {
  GPUNodeConfiguration *GPUConf =
      (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);
  std::vector<std::pair<
      GPUNodeConfiguration::TENSOR_OP,
      std::vector<std::pair<GPUNodeConfiguration::APPROX, int>>>>
      &ApproxChoices = GPUConf->getApproxChoices();
  // Approximation choices must be for a projectiveT operation
  CUSTOM_ASSERT(
      ApproxChoices.size() == 1 &&
      ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::PROJECTIVE_T &&
      "Invalid configuration generated for tensor projectiveT "
      "wrapper operation");
  return handleTensorProjectiveTApproximationTuples(
      ApproxChoices[0].second, input, transformation);
}

void *wrapper_tensorMap1(const char *hpvm_node_id, int func, void *input) {
  GPUNodeConfiguration *GPUConf =
      (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);
  std::vector<std::pair<
      GPUNodeConfiguration::TENSOR_OP,
      std::vector<std::pair<GPUNodeConfiguration::APPROX, int>>>>
      &ApproxChoices = GPUConf->getApproxChoices();
  // Approximation choices must be for a map1 operation
  CUSTOM_ASSERT(
      ApproxChoices.size() == 1 &&
      ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::MAP1 &&
      "Invalid configuration generated for tensor map1 wrapper operation");
  return handleTensorMap1ApproximationTuples(
      ApproxChoices[0].second, (MathOp)func, input);
}

void *wrapper_tensorMap2(
    const char *hpvm_node_id, int func, void *input1, void *input2) {
  GPUNodeConfiguration *GPUConf =
      (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);
  std::vector<std::pair<
      GPUNodeConfiguration::TENSOR_OP,
      std::vector<std::pair<GPUNodeConfiguration::APPROX, int>>>>
      &ApproxChoices = GPUConf->getApproxChoices();
  // Approximation choices must be for a map2 operation
  CUSTOM_ASSERT(
      ApproxChoices.size() == 1 &&
      ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::MAP2 &&
      "Invalid configuration generated for tensor map2 wrapper operation");
  return handleTensorMap2ApproximationTuples(
      ApproxChoices[0].second, (MathOp)func, input1, input2);
}

void *wrapper_tensorMap3(
    const char *hpvm_node_id, int func, void *input1, void *input2,
    void *input3) {
  GPUNodeConfiguration *GPUConf =
      (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);
  std::vector<std::pair<
      GPUNodeConfiguration::TENSOR_OP,
      std::vector<std::pair<GPUNodeConfiguration::APPROX, int>>>>
      &ApproxChoices = GPUConf->getApproxChoices();
  // Approximation choices must be for a map3 operation
  CUSTOM_ASSERT(
      ApproxChoices.size() == 1 &&
      ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::MAP3 &&
      "Invalid configuration generated for tensor map3 wrapper operation");
  return handleTensorMap3ApproximationTuples(
      ApproxChoices[0].second, (MathOp)func, input1, input2, input3);
}

// Tentative
void *wrapper_tensorStencil(const char *hpvm_node_id, void *input) {
  ERROR("Stencil operation currently unsupported.\n");
  abort();
}

void *wrapper_tensorCosineT(const char *hpvm_node_id, void *input) {
  ERROR("CosineT operation currently unsupported.\n");
  abort();
}
