#include <algorithm>
#include <cmath>
#include <cstring>
#include <experimental/filesystem>
#include <numeric>
#include <sstream>
#include <string>

#include "debug.h"
#include "device_math.h"
#include "functional/common.h"
#include "img_tensor_runtime.h"
#include "img_tensor_utils.h"
#include "tensor_utils.h"

// Image I/O utilities
#define STB_IMAGE_IMPLEMENTATION
#define STB_IMAGE_WRITE_IMPLEMENTATION

#include "image/stb_image.h"
#include "image/stb_image_write.h"

static inline uint8_t *float_to_uint8(const float *fl, size_t count) {
  auto *ret = new uint8_t[count];
  float max_v = *std::max_element(fl, fl + count),
        min_v = *std::min_element(fl, fl + count);
  if (max_v - min_v < 1e-3) {
    for (size_t i = 0; i < count; i++)
      ret[i] = 0;
  } else {
    float frac = 255 / (max_v - min_v);
    for (size_t i = 0; i < count; i++)
      ret[i] = uint8_t(frac * (fl[i] - min_v));
  }
  return ret;
}

static inline float *uint8_to_float(const uint8_t *ui, size_t len) {
  auto *ret = new float[len];
  for (size_t i = 0; i < len; i++)
    ret[i] = float(ui[i]) / 255;
  return ret;
}

static Tensor *to_nhwc(Tensor *t) {
  if (t->data_format == CUDNN_TENSOR_NHWC) {
    DEBUG("Tensor already in NHWC format, no conversion needed\n");
    return t;
  } else if (t->data_format != CUDNN_TENSOR_NCHW) {
    ERROR("Unknown tensor format: %s\n", std::to_string(t->data_format));
  } else {
    DEBUG("Converting to NHWC format\n");
  }

  size_t *dim_arr = t->dims.dim_sizes;
  size_t n = dim_arr[0], c = dim_arr[1], h = dim_arr[2], w = dim_arr[3];
  auto *out_tensor =
      (Tensor *)create4DTensor(t->data_type, CUDNN_TENSOR_NHWC, n, h, w, c);
  size_t nhwc_offset = 0;
  size_t element_size = getTypeSize(t->data_type);
  char *out_data = (char *)(out_tensor->host_data),
       *in_data = (char *)(t->host_data);
  for (int n0 = 0; n0 < n; n0++)
    for (int h0 = 0; h0 < h; h0++)
      for (int w0 = 0; w0 < w; w0++)
        for (int c0 = 0; c0 < c; c0++) {
          size_t nc = n0 * c + c0, nch = nc * h + h0, nchw_idx = nch * w + w0,
                 nchw_offset = nchw_idx * element_size;
          std::memcpy(out_data + nhwc_offset, in_data + nchw_offset,
                      element_size);
          nhwc_offset += element_size;
        }
  return out_tensor;
}

static Tensor *to_nchw(Tensor *t) {
  if (t->data_format == CUDNN_TENSOR_NCHW) {
    DEBUG("Tensor already in NCHW format, no conversion needed\n");
    return t;
  } else if (t->data_format != CUDNN_TENSOR_NHWC) {
    ERROR("Unknown tensor format: %s\n", std::to_string(t->data_format));
  } else {
    DEBUG("Converting to NCHW format\n");
  }
  size_t *dim_arr = t->dims.dim_sizes;
  size_t n = dim_arr[0], h = dim_arr[1], w = dim_arr[2], c = dim_arr[3];
  Tensor *out_tensor =
      (Tensor *)create4DTensor(t->data_type, CUDNN_TENSOR_NCHW, n, c, h, w);
  size_t nchw_offset = 0;
  size_t element_size = getTypeSize(t->data_type);
  char *out_data = (char *)(out_tensor->host_data),
       *in_data = (char *)(t->host_data);
  for (int n0 = 0; n0 < n; n0++)
    for (int c0 = 0; c0 < c; c0++)
      for (int h0 = 0; h0 < h; h0++)
        for (int w0 = 0; w0 < w; w0++) {
          size_t nh = n0 * h + h0, nhw = nh * w + w0, nhwc_idx = nhw * c + c0,
                 nhwc_offset = nhwc_idx * element_size;
          std::memcpy(out_data + nchw_offset, in_data + nhwc_offset,
                      element_size);
          nchw_offset += element_size;
        }
  return out_tensor;
}

namespace fs = std::experimental::filesystem;

// List all files in a folder.
static inline std::vector<std::string> listFiles(const std::string &folder) {
  std::vector<std::string> ret;
  for (const auto &entry : fs::directory_iterator(folder))
    ret.push_back(entry.path().string());
  std::sort(ret.begin(), ret.end());
  return ret;
}

// return in[start:start+count]
template <typename T>
std::vector<T> sliceVector(const std::vector<T> &in, size_t start,
                           size_t count) {
  auto slice_begin = in.begin() + start;
  if (slice_begin > in.end())
    slice_begin = in.end();
  auto slice_end = count == std::string::npos ? in.end() : slice_begin + count;
  if (slice_end > in.end())
    slice_end = in.end();
  return std::vector<T>(slice_begin, slice_end);
}

// Read an image dataset from a folder with each image as a file.
Tensor *readDataSet(const char *path, size_t start, size_t count,
                    size_t n_color) {
  INFO("Loading image dataset from path %s\n", path);
  std::vector<std::string> filenames =
      sliceVector(listFiles(path), start, count);
  if (filenames.empty()) {
    INFO("Folder is empty or slice is empty\n");
    return nullptr;
  }

  auto *first_image = (Tensor *)loadAsImage(filenames[0].c_str(), n_color);
  std::vector<size_t> sizes = ::sizes(first_image);
  size_t h = sizes[2], w = sizes[3];
  DEBUG("Loading shape: (%lu, %lu, %lu, %lu)\n", filenames.size(), n_color, h,
        w);
  auto *batch = (Tensor *)create4DTensor(CUDNN_DATA_FLOAT, CUDNN_TENSOR_NHWC,
                                         filenames.size(), h, w, n_color);
  size_t n_floats = n_color * h * w;
  auto *base_data = (float *)batch->host_data;
  for (const auto &path : filenames) {
    int x, y, n; // x = width, y = height, n = # 8-bit components per pixel
    uint8_t *data = stbi_load(path.c_str(), &x, &y, &n, n_color);
    if (data == nullptr)
      ERROR("Image load failed\n");
    if (x != h || y != w) {
      std::ostringstream os;
      os << "Image file " << path << " have different shape (" << x << ", " << y
         << ")";
      ERROR("%s\n", os.str());
    }
    float *converted = uint8_to_float(data, n_floats);
    stbi_image_free(data);
    std::memcpy(base_data, converted, n_floats * sizeof(float));
    delete[] converted;
    base_data += n_floats;
  }
  auto *nchw_batch = to_nchw(batch);
  DEBUG("Loaded all images.\n");
  return nchw_batch;
}

// Convert complex-domain image to float valued image.
static Tensor *complexToFloat(Tensor *batch) {
  convertAndGetGPUData<float2>(batch); // Convert to float2
  deviceToHostCopy(batch);
  auto *in_data = (float2 *)batch->host_data;
  size_t n_elem = batch->num_elems;
  std::vector<float> magnitudes(n_elem, 0.0f);
  for (size_t i = 0; i < batch->num_elems; i++) {
    magnitudes[i] = hypot(in_data[i].x, in_data[i].y);
  }

  size_t *dims = batch->dims.dim_sizes;
  auto *ret = (Tensor *)create4DTensor(float_type, batch->data_format, dims[0],
                                       dims[1], dims[2], dims[3]);
  auto *out_data = (float *)ret->host_data;
  for (size_t i = 0; i < magnitudes.size(); i++) {
    float f = magnitudes[i];
    out_data[i] = f > 1.0f ? log(f) : 0;
  }
  return ret;
}

// Save an image tensor image-by-image to a folder.
void saveDataSet(const char *path, Tensor *batch, size_t start_idx,
                 size_t write_n) {
  INFO("Saving image dataset to path %s\n", path);
  Tensor *float_batch = batch;
  if (batch->data_type == float2_type || batch->data_type == half2_type)
    float_batch = complexToFloat(float_batch); // Already copied
  else {
    DEBUG("Copying to CPU before printing\n");
    convertAndGetGPUData<float>(float_batch);
    deviceToHostCopy(float_batch);
  }
  Tensor *converted_batch = float_batch;
  if (converted_batch->data_format == CUDNN_TENSOR_NCHW) {
    DEBUG("Copy-converting to NHWC format\n");
    converted_batch = to_nhwc(converted_batch);
  }
  std::vector<size_t> sizes = ::sizes(converted_batch);
  size_t h = sizes[1], w = sizes[2], c = sizes[3];
  auto *base_data = (float *)converted_batch->host_data;
  if (write_n == 0)
    write_n = sizes[0];
  else
    write_n = std::min(write_n, sizes[0]);
  for (size_t i = start_idx; i < start_idx + write_n; i++) {
    std::string name = path;
    name += "/";
    std::string number = std::to_string(i);
    // FIXME: pad to 6 digits. Ordering will break when we have more than 1M
    // files.
    number = std::string(6 - number.length(), '0') + number;
    name += number + ".png";

    uint8_t *ldr_data = float_to_uint8(base_data, h * w * c);
    if (!stbi_write_png(name.c_str(), w, h, c, ldr_data, 0))
      ERROR("Write file failed\n");
    delete[] ldr_data;

    base_data += h * w * c;
  }
}

// Load 1 file as an image into a tensor.
void *loadAsImage(const char *filename, size_t n_color) {
  INFO("Loading image from path=%s\n", filename);
  int x, y, n; // x = width, y = height, n = # 8-bit components per pixel
  uint8_t *data = stbi_load(filename, &x, &y, &n, n_color);
  if (data == nullptr)
    ERROR("Image load failed\n");
  float *converted = uint8_to_float(data, x * y * n);
  DEBUG("Loading shape: (1, %lu, %lu, %lu)(NHWC)\n", y, x, n_color);
  auto *image =
      (Tensor *)create4DTensor(CUDNN_DATA_FLOAT, CUDNN_TENSOR_NHWC, 1, y, x, n);
  std::memcpy(image->host_data, converted, x * y * n * sizeof(float));
  auto *nchw_image = to_nchw(image);
  stbi_image_free(data);
  return nchw_image;
}

// Save 1 tensor as an image into a file.
void saveToImage(const char *filename, Tensor *tensor) {
  INFO("Saving image data to path=%s\n", filename);
  deviceToHostCopy(tensor);
  Tensor *converted_tensor = tensor;
  if (tensor->data_format == CUDNN_TENSOR_NCHW) {
    DEBUG("Copy-converting to NHWC format\n");
    converted_tensor = to_nhwc(tensor);
  }
  auto *hdr_data = (float *)converted_tensor->host_data;
  size_t *dims = converted_tensor->dims.dim_sizes;
  size_t w = dims[2], h = dims[1], c = dims[3];
  uint8_t *ldr = float_to_uint8(hdr_data, w * h * c);
  stbi_write_png(filename, w, h, c, ldr, 0);
  delete[] ldr;
}

// Make a conv2d filter from 2-dim data.
void *createFilterFromData(int data_type, void *data, size_t w, size_t h,
                           size_t n_chan) {
  DEBUG("Creating filter from data\n");
  auto *tensor =
      (Tensor *)create4DTensor(data_type, CUDNN_TENSOR_NCHW, n_chan, 1, h, w);
  char *tensor_data;
  if (data_type == CUDNN_DATA_HALF || data_type == CUDNN_DATA_FLOAT)
    tensor_data = (char *)tensor->host_data;
  else {
    ERROR("Data type unsupported as filter\n");
  }
  size_t channel_sz = tensor->size_in_bytes / n_chan;
  for (size_t i = 0; i < n_chan; i++, tensor_data += channel_sz) {
    std::memcpy(tensor_data, data, channel_sz);
  }
  return tensor;
}

// Normalize an image tensor.
static void *normalize(void *image) {
  auto *max_1D = tensorReduce(image, 2, MathOp::Max);
  auto *max = tensorReduce(max_1D, 3, MathOp::Max);
  auto *img_norm = tensorMap2(MathOp::Div, image, max);
  freeTensor(max_1D);
  freeTensor(max);
  return img_norm;
}

float compute_mean(float *arr, int left, int right) {
  float sum = 0;
  for (int i = left; i < right; i++) {
    sum += arr[i];
  }
  return sum / (right - left);
}

float compute_variance(float *arr, int left, int right, float mean) {
  float sum = 0;
  for (int i = left; i < right; i++) {
    sum += (arr[i] - mean) * (arr[i] - mean);
  }
  return sum / (right - left - 1);
}

float compute_covariance(float *x, float *y, int left, int right, float x_mean,
                         float y_mean) {
  float sum = 0;
  for (int i = left; i < right; i++) {
    sum += (x[i] - x_mean) * (y[i] - y_mean);
  }
  return sum / (right - left - 1);
}

std::vector<float> SSIM(void *lhs_ptr, void *rhs_ptr) {
  auto *lhs = (Tensor *)lhs_ptr, *rhs = (Tensor *)rhs_ptr;

  lhs = (Tensor *)normalize(lhs);
  rhs = (Tensor *)normalize(rhs);

  convertToFP32(lhs);
  convertToFP32(rhs);
  deviceToHostCopy(lhs);
  deviceToHostCopy(rhs);

  float L = 1.0;
  float K1 = 0.01;
  float K2 = 0.03;
  float C1 = (K1 * L) * (K1 * L);
  float C2 = (K2 * L) * (K2 * L);

  int n = lhs->dims.dim_sizes[0];
  int c = lhs->dims.dim_sizes[1];
  int h = lhs->dims.dim_sizes[2];
  int w = lhs->dims.dim_sizes[3];

  float *lhs_arr = (float *)lhs->host_data;
  float *rhs_arr = (float *)rhs->host_data;

  std::vector<float> scores;
  for (int i = 0; i < n; i++) {
    int left = i * c * h * w;
    int right = (i + 1) * c * h * w;

    float x_mean = compute_mean(lhs_arr, left, right);
    float y_mean = compute_mean(rhs_arr, left, right);
    float x_var = compute_variance(lhs_arr, left, right, x_mean);
    float y_var = compute_variance(rhs_arr, left, right, y_mean);
    float covariance =
        compute_covariance(lhs_arr, rhs_arr, left, right, x_mean, y_mean);

    scores.push_back(
        ((2 * x_mean * y_mean + C1) * (2 * covariance + C2)) /
        ((x_mean * x_mean + y_mean * y_mean + C1) * (x_var + y_var + C2)));
  }
  return scores;
}

std::vector<float> PSNR(void *gold_ptr, void *approx_ptr) {
  auto *gold_tensor = (Tensor *)gold_ptr, *approx_tensor = (Tensor *)approx_ptr;
  convertToFP32(gold_tensor);
  convertToFP32(approx_tensor);

  size_t *dim_sizes = gold_tensor->dims.dim_sizes;
  size_t batch_dim = dim_sizes[0];
  size_t image_size = dim_sizes[1] * dim_sizes[2] * dim_sizes[3];
  float image_size_f = image_size;
  DEBUG("batch_dim = %lu, image_size = %lu\n", batch_dim, image_size);
  auto *image_size_tensor =
      (Tensor *)create4DTensor(CUDNN_DATA_FLOAT, CUDNN_TENSOR_NCHW, 1, 1, 1, 1);
  std::memcpy(image_size_tensor->host_data, &image_size_f, sizeof(float));

  gold_tensor = (Tensor *)normalize(gold_tensor);
  approx_tensor = (Tensor *)normalize(approx_tensor);
  auto *diff = tensorMap2(MathOp::Sub, gold_tensor, approx_tensor);
  auto *diffsqr = tensorMap2(MathOp::Mul, diff, diff);
  auto *mse_sum_2d = tensorReduce(diffsqr, 3, MathOp::Add);
  auto *mse_sum_1d = tensorReduce(mse_sum_2d, 2, MathOp::Add);
  auto *mse_sum = tensorReduce(mse_sum_1d, 1, MathOp::Add);
  auto *mse_avg = tensorMap2(MathOp::Div, mse_sum, image_size_tensor);
  auto *psnr_val = (Tensor *)tensorMap1(MathOp::PSNR, mse_avg);
  deviceToHostCopy(psnr_val);

  auto *float_data = (float *)psnr_val->host_data;
  return std::vector<float>(float_data, float_data + batch_dim);
}

float violationRate(const std::vector<float> &values, float threshold,
                    bool higher_better) {
  if (values.empty())
    return 0.0f;
  size_t violation = 0;
  for (float v : values) {
    if (std::isnan(v))
      ++violation;
    if ((higher_better && v < threshold) || (!higher_better && v > threshold))
      ++violation;
  }
  return (float)violation / values.size();
}

float mean(const std::vector<float> &values) {
  std::vector<float> non_nan;
  for (float f : values)
    if (!std::isnan(f))
      non_nan.push_back(f);
  if (non_nan.empty())
    return 0.0f;
  return std::accumulate(non_nan.begin(), non_nan.end(), 0.0f, std::plus<>()) /
         (float)non_nan.size();
}

void *sliceTensorInBatch(void *whole, size_t start, size_t end) {
  auto *whole_tensor = (Tensor *)whole;
  size_t *dim_sizes = whole_tensor->dims.dim_sizes;
  auto *output =
      (Tensor *)create4DTensor(CUDNN_DATA_FLOAT, CUDNN_TENSOR_NCHW, end - start,
                               dim_sizes[1], dim_sizes[2], dim_sizes[3]);
  size_t single_size = dim_sizes[1] * dim_sizes[2] * dim_sizes[3];
  auto *in_data = (float *)(whole_tensor->host_data) + start * single_size;
  memcpy(output->host_data, in_data, (end - start) * single_size);
  return output;
}

void reshape(void *t, const std::vector<size_t> &shape) {
  auto *tensor = (Tensor *)t;
  size_t in_n = num_elems(tensor), out_n = num_elems(shape);
  if (in_n != out_n)
    ERROR("Reshaping cannot change number of elements\n");
  tensor->dims.num_dims = shape.size();
  free(tensor->dims.dim_sizes);
  tensor->dims.dim_sizes = (size_t *)malloc(sizeof(size_t) * shape.size());
  std::copy(shape.begin(), shape.end(), tensor->dims.dim_sizes);
  set4DTensorDescriptor(tensor, tensor->data_format, shape[0], shape[1],
                        shape[2], shape[3]);
}
