

#include "tensor_utils.h"
#include "fp16_gemm.h"
#include "debug.h"
#include "global_data.h"
#include "profiling.h"


extern "C"{



__global__
void depthwise_conv(float* const __restrict__ y,
const float* const __restrict__ x,
const float* const __restrict__ w,
const int B, const int M,
const int H, const int W, const int KH,
const int KW, const int H_out, const int W_out,
const int H_pad, const int W_pad,
const int H_stride, const int W_stride, const int start_batch)
{

#define y4d(i3, i2, i1, i0) y[(i3) * (M * H_out * W_out) + (i2) * (H_out * W_out) + (i1) * (W_out) + i0]
#define x4d(i3, i2, i1, i0) x[(i3) * (M * H * W) + (i2) * (H * W) + (i1) * (W) + i0]

const int num = 1;

const int b = num * blockIdx.x + start_batch;
const int m = blockIdx.y; //current filter/channel

const int tx = threadIdx.x;

const int start_h = (threadIdx.x / W_out) * H_stride - H_pad;
const int start_w = (threadIdx.x % W_out) * W_stride - W_pad;

float C[num] = { 0 };

const float* weights = &w[m * KH * KW];

for (int k = 0; k < KH * KW; k++) {
int p = k / KW;
int q = k % KW;

#pragma unroll
for (int i = 0; i < num; i++) {
if (start_h + p > -1 && start_h + p < H &&
start_w + q > -1 && start_w + q < W) {

C[i] += x4d(b + i, m, start_h + p, start_w + q) * weights[k];
}

}
}

#pragma unroll
for (int i = 0; i < num; i++) {
if(b + i < B)
y4d(b + i, m, 0, tx) = C[i];

}


#undef y4d 
#undef x4d
}


__global__
void depthwise_convNew(float* const __restrict__ y,
const float* const __restrict__ x,
const float* const __restrict__ w,
const int B, const int M,
const int H, const int W, const int KH,
const int KW, const int H_out, const int W_out,
const int H_pad, const int W_pad,
const int H_stride, const int W_stride)
{

#define y4d(i3, i2, i1, i0) y[(i3) * (M * H_out * W_out) + (i2) * (H_out * W_out) + (i1) * (W_out) + i0]
#define x4d(i3, i2, i1, i0) x[(i3) * (M * H * W) + (i2) * (H * W) + (i1) * (W) + i0]

const int num = 12;

const int b = num * blockIdx.x;
const int m = (blockIdx.y * blockDim.x  + threadIdx.x)/ (H_out * W_out); 

const int tx = (blockIdx.y * blockDim.x  + threadIdx.x) % (H_out * W_out);

const int start_h = (tx / W_out) * H_stride - H_pad;
const int start_w = (tx % W_out) * W_stride - W_pad;

float C[num] = { 0 };

const float* weights = &w[m * KH * KW];

for (int k = 0; k < KH * KW; k++) {
int p = k / KW;
int q = k % KW;

if (start_h + p > -1 && start_h + p < H &&
start_w + q > -1 && start_w + q < W) {

#pragma unroll
for (int i = 0; i < num; i++) {
if(b + i < B)
C[i] += x4d(b + i, m, start_h + p, start_w + q) * weights[k];
}

}
}

#pragma unroll
for (int i = 0; i < num; i++) {
if(b + i < B)
y4d(b + i, m, 0, tx) = C[i];

}

#undef y4d 
#undef x4d
}




__global__ void depthwise_convNew8_half(__half* const __restrict__ y,
					const __half* const __restrict__ x,
					const __half* const __restrict__ w,
					const int B, const int M,
					const int H, const int W, const int KH,
					const int KW, const int H_out, const int W_out,
					const int H_pad, const int W_pad,
					const int H_stride, const int W_stride)
{

  #define y4d(i3, i2, i1, i0) y[(i3) * (M * H_out * W_out) + (i2) * (H_out * W_out) + (i1) * (W_out) + i0]
  #define x4d(i3, i2, i1, i0) x[(i3) * (M * H * W) + (i2) * (H * W) + (i1) * (W) + i0]

  const int num = 8;

  const int b = num * blockIdx.x;
  const int m = (blockIdx.y * blockDim.x  + threadIdx.x)/ (H_out * W_out);
	
  if(m < M){
    const int tx = (blockIdx.y * blockDim.x  + threadIdx.x) % (H_out * W_out);

    const int start_h = (tx / W_out) * H_stride - H_pad;
    const int start_w = (tx % W_out) * W_stride - W_pad;

    __half c0 = 0;
    __half c1 = 0;
    __half c2 = 0;
    __half c3 = 0;
    __half c4 = 0;
    __half c5 = 0;
    __half c6 = 0;
    __half c7 = 0;
	
    const __half* weights = &w[m * KH * KW];

    for (int k = 0; k < KH * KW; k++) {
      int p = k / KW;
      int q = k % KW;

      if (start_h + p > -1 && start_h + p < H &&
	  start_w + q > -1 && start_w + q < W) {

	c0 = __hfma(x4d(b, m, start_h + p, start_w + q), weights[k], c0);
	if(b + 1 < B)
	  c1 = __hfma(x4d(b + 1, m, start_h + p, start_w + q), weights[k], c1);
	if(b + 2 < B)
	  c2 = __hfma(x4d(b + 2, m, start_h + p, start_w + q), weights[k], c2);
	if(b + 3 < B)
	  c3 = __hfma(x4d(b + 3, m, start_h + p, start_w + q), weights[k], c3);
	if(b + 4 < B)
	  c4 = __hfma(x4d(b + 4, m, start_h + p, start_w + q), weights[k], c4);
	if(b + 5 < B)
	  c5 = __hfma(x4d(b + 5, m, start_h + p, start_w + q), weights[k], c5);
	if(b + 6 < B)
	  c6 = __hfma(x4d(b + 6, m, start_h + p, start_w + q), weights[k], c6);
	if(b + 7 < B)
	  c7 = __hfma(x4d(b + 7, m, start_h + p, start_w + q), weights[k], c7);
    

      }
    }

    y4d(b, m, 0, tx) = c0;	
    if(b + 1 < B)
      y4d(b + 1, m, 0, tx) = c1;
    if(b + 2 < B)
      y4d(b + 2, m, 0, tx) = c2;
    if(b + 3 < B)
      y4d(b + 3, m, 0, tx) = c3;
    if(b + 4 < B)
      y4d(b + 4, m, 0, tx) = c4;
    if(b + 5 < B)
      y4d(b + 5, m, 0, tx) = c5;
    if(b + 6 < B)
      y4d(b + 6, m, 0, tx) = c6;
    if(b + 7 < B)
      y4d(b + 7, m, 0, tx) = c7;
  }
	
  #undef y4d 
  #undef x4d
}

__global__ void depthwise_convNew8_half1(__half* const __restrict__ y,
					const __half* const __restrict__ x,
					const __half* const __restrict__ w,
					const int B, const int M,
					const int H, const int W, const int KH,
					const int KW, const int H_out, const int W_out,
					const int H_pad, const int W_pad,
					const int H_stride, const int W_stride)
{

  #define y4d(i3, i2, i1, i0) y[(i3) * (M * H_out * W_out) + (i2) * (H_out * W_out) + (i1) * (W_out) + i0]
  #define x4d(i3, i2, i1, i0) x[(i3) * (M * H * W) + (i2) * (H * W) + (i1) * (W) + i0]

  const int num = 8;

  const int b = num * blockIdx.x;
  const int m = (blockIdx.y * blockDim.x  + threadIdx.x)/ (H_out * W_out);
	
  if(m < M){
    const int tx = (blockIdx.y * blockDim.x  + threadIdx.x) % (H_out * W_out);

    const int start_h = (tx / W_out) * H_stride - H_pad;
    const int start_w = (tx % W_out) * W_stride - W_pad;

    __half c0 = 0;
    __half c1 = 0;
    __half c2 = 0;
    __half c3 = 0;
    __half c4 = 0;
    __half c5 = 0;
    __half c6 = 0;
    __half c7 = 0;
	
    const __half* weights = &w[m * KH * KW];

    for (int k = 0; k < KH * KW; k++) {
      int p = k / KW;
      int q = k % KW;

      if (start_h + p > -1 && start_h + p < H &&
	  start_w + q > -1 && start_w + q < W) {

	c0 = __hfma(x4d(b, m, start_h + p, start_w + q), weights[k], c0);
      }
    }

    if(b + 1 < B){
      for (int k = 0; k < KH * KW; k++) {
	int p = k / KW;
	int q = k % KW;

	if (start_h + p > -1 && start_h + p < H &&
	    start_w + q > -1 && start_w + q < W) {

	  c1 = __hfma(x4d(b + 1, m, start_h + p, start_w + q), weights[k], c1);
	}
      }
    }

    if(b + 2 < B){
      for (int k = 0; k < KH * KW; k++) {
	int p = k / KW;
	int q = k % KW;

	if (start_h + p > -1 && start_h + p < H &&
	    start_w + q > -1 && start_w + q < W) {

	  c2 = __hfma(x4d(b + 2, m, start_h + p, start_w + q), weights[k], c2);
	}
      }
    }

    if(b + 3 < B){
      for (int k = 0; k < KH * KW; k++) {
	int p = k / KW;
	int q = k % KW;

	if (start_h + p > -1 && start_h + p < H &&
	    start_w + q > -1 && start_w + q < W) {

	  c3 = __hfma(x4d(b + 3, m, start_h + p, start_w + q), weights[k], c3);
	}
      }
    }

    if(b + 4 < B){
      for (int k = 0; k < KH * KW; k++) {
	int p = k / KW;
	int q = k % KW;

	if (start_h + p > -1 && start_h + p < H &&
	    start_w + q > -1 && start_w + q < W) {

	  c4 = __hfma(x4d(b + 4, m, start_h + p, start_w + q), weights[k], c4);
	}
      }
    }

    if(b + 5 < B){
      for (int k = 0; k < KH * KW; k++) {
	int p = k / KW;
	int q = k % KW;

	if (start_h + p > -1 && start_h + p < H &&
	    start_w + q > -1 && start_w + q < W) {

	  c5 = __hfma(x4d(b + 5, m, start_h + p, start_w + q), weights[k], c5);
	}
      }
    }

    if(b + 6 < B){
      for (int k = 0; k < KH * KW; k++) {
	int p = k / KW;
	int q = k % KW;

	if (start_h + p > -1 && start_h + p < H &&
	    start_w + q > -1 && start_w + q < W) {

	  c6 = __hfma(x4d(b + 6, m, start_h + p, start_w + q), weights[k], c6);
	}
      }
    }

    if(b + 7 < B){
      for (int k = 0; k < KH * KW; k++) {
	int p = k / KW;
	int q = k % KW;

	if (start_h + p > -1 && start_h + p < H &&
	    start_w + q > -1 && start_w + q < W) {

	  c7 = __hfma(x4d(b + 7, m, start_h + p, start_w + q), weights[k], c7);
	}
      }
    }

    

    y4d(b, m, 0, tx) = c0;	
    if(b + 1 < B)
      y4d(b + 1, m, 0, tx) = c1;
    if(b + 2 < B)
      y4d(b + 2, m, 0, tx) = c2;
    if(b + 3 < B)
      y4d(b + 3, m, 0, tx) = c3;
    if(b + 4 < B)
      y4d(b + 4, m, 0, tx) = c4;
    if(b + 5 < B)
      y4d(b + 5, m, 0, tx) = c5;
    if(b + 6 < B)
      y4d(b + 6, m, 0, tx) = c6;
    if(b + 7 < B)
      y4d(b + 7, m, 0, tx) = c7;
  }
	
  #undef y4d 
  #undef x4d
}








__global__ void depthwise_convNew12(float* const __restrict__ y,
				    const float* const __restrict__ x,
				    const float* const __restrict__ w,
				    const int B, const int M,
				    const int H, const int W, const int KH,
				    const int KW, const int H_out, const int W_out,
				    const int H_pad, const int W_pad,
				    const int H_stride, const int W_stride)
{

  #define y4d(i3, i2, i1, i0) y[(i3) * (M * H_out * W_out) + (i2) * (H_out * W_out) + (i1) * (W_out) + i0]
  #define x4d(i3, i2, i1, i0) x[(i3) * (M * H * W) + (i2) * (H * W) + (i1) * (W) + i0]

  const int num = 12;

  const int b = num * blockIdx.x;
  const int m = (blockIdx.y * blockDim.x  + threadIdx.x)/ (H_out * W_out);
	
  if(m < M){
    const int tx = (blockIdx.y * blockDim.x  + threadIdx.x) % (H_out * W_out);

    const int start_h = (tx / W_out) * H_stride - H_pad;
    const int start_w = (tx % W_out) * W_stride - W_pad;

    float c0 = 0;
    float c1 = 0;
    float c2 = 0;
    float c3 = 0;
    float c4 = 0;
    float c5 = 0;
    float c6 = 0;
    float c7 = 0;
    float c8 = 0;
    float c9 = 0;
    float c10 = 0;
    float c11 = 0;
	
    const float* weights = &w[m * KH * KW];

    for (int k = 0; k < KH * KW; k++) {
      int p = k / KW;
      int q = k % KW;

      if (start_h + p > -1 && start_h + p < H &&
	  start_w + q > -1 && start_w + q < W) {

	c0 += x4d(b, m, start_h + p, start_w + q) * weights[k];
	if(b + 1 < B)
	  c1 += x4d(b + 1, m, start_h + p, start_w + q) * weights[k];
	if(b + 2 < B)
	  c2 += x4d(b + 2, m, start_h + p, start_w + q) * weights[k];
	if(b + 3 < B)
	  c3 += x4d(b + 3, m, start_h + p, start_w + q) * weights[k];
	if(b + 4 < B)
	  c4 += x4d(b + 4, m, start_h + p, start_w + q) * weights[k];
	if(b + 5 < B)
	  c5 += x4d(b + 5, m, start_h + p, start_w + q) * weights[k];
	if(b + 6 < B)
	  c6 += x4d(b + 6, m, start_h + p, start_w + q) * weights[k];
	if(b + 7 < B)
	  c7 += x4d(b + 7, m, start_h + p, start_w + q) * weights[k];
	if(b + 8 < B)
	  c8 += x4d(b + 8, m, start_h + p, start_w + q) * weights[k];
	if(b + 9 < B)
	  c9 += x4d(b + 9, m, start_h + p, start_w + q) * weights[k];
	if(b + 10 < B)
	  c10 += x4d(b + 10, m, start_h + p, start_w + q) * weights[k];
	if(b + 11 < B)
	  c11 += x4d(b + 11, m, start_h + p, start_w + q) * weights[k];
    

      }
    }

    y4d(b, m, 0, tx) = c0;	
    if(b + 1 < B)
      y4d(b + 1, m, 0, tx) = c1;
    if(b + 2 < B)
      y4d(b + 2, m, 0, tx) = c2;
    if(b + 3 < B)
      y4d(b + 3, m, 0, tx) = c3;
    if(b + 4 < B)
      y4d(b + 4, m, 0, tx) = c4;
    if(b + 5 < B)
      y4d(b + 5, m, 0, tx) = c5;
    if(b + 6 < B)
      y4d(b + 6, m, 0, tx) = c6;
    if(b + 7 < B)
      y4d(b + 7, m, 0, tx) = c7;
    if(b + 8 < B)
      y4d(b + 8, m, 0, tx) = c8;
    if(b + 9 < B)
      y4d(b + 9, m, 0, tx) = c9;
    if(b + 10 < B)
      y4d(b + 10, m, 0, tx) = c10;
    if(b + 11 < B)
      y4d(b + 11, m, 0, tx) = c11;
	
  }
	
  #undef y4d 
  #undef x4d
}


__global__ void depthwise_convNew12_half(__half* const __restrict__ y,
				    const __half* const __restrict__ x,
				    const __half* const __restrict__ w,
				    const int B, const int M,
				    const int H, const int W, const int KH,
				    const int KW, const int H_out, const int W_out,
				    const int H_pad, const int W_pad,
				    const int H_stride, const int W_stride)
{

  #define y4d(i3, i2, i1, i0) y[(i3) * (M * H_out * W_out) + (i2) * (H_out * W_out) + (i1) * (W_out) + i0]
  #define x4d(i3, i2, i1, i0) x[(i3) * (M * H * W) + (i2) * (H * W) + (i1) * (W) + i0]

  const int num = 12;

  const int b = num * blockIdx.x;
  const int m = (blockIdx.y * blockDim.x  + threadIdx.x)/ (H_out * W_out);
	
  if(m < M){
    const int tx = (blockIdx.y * blockDim.x  + threadIdx.x) % (H_out * W_out);

    const int start_h = (tx / W_out) * H_stride - H_pad;
    const int start_w = (tx % W_out) * W_stride - W_pad;

    __half c0 = 0;
    __half c1 = 0;
    __half c2 = 0;
    __half c3 = 0;
    __half c4 = 0;
    __half c5 = 0;
    __half c6 = 0;
    __half c7 = 0;
    __half c8 = 0;
    __half c9 = 0;
    __half c10 = 0;
    __half c11 = 0;
	
    const __half* weights = &w[m * KH * KW];

    for (int k = 0; k < KH * KW; k++) {
      int p = k / KW;
      int q = k % KW;

      if (start_h + p > -1 && start_h + p < H &&
	  start_w + q > -1 && start_w + q < W) {

	c0 = __hfma(x4d(b, m, start_h + p, start_w + q), weights[k], c0);
	if(b + 1 < B)
	  c1 = __hfma(x4d(b + 1, m, start_h + p, start_w + q), weights[k], c1);
	if(b + 2 < B)
	  c2 = __hfma(x4d(b + 2, m, start_h + p, start_w + q), weights[k], c2);
	if(b + 3 < B)
	  c3 = __hfma(x4d(b + 3, m, start_h + p, start_w + q), weights[k], c3);
	if(b + 4 < B)
	  c4 = __hfma(x4d(b + 4, m, start_h + p, start_w + q), weights[k], c4);
	if(b + 5 < B)
	  c5 = __hfma(x4d(b + 5, m, start_h + p, start_w + q), weights[k], c5);
	if(b + 6 < B)
	  c6 = __hfma(x4d(b + 6, m, start_h + p, start_w + q), weights[k], c6);
	if(b + 7 < B)
	  c7 = __hfma(x4d(b + 7, m, start_h + p, start_w + q), weights[k], c7);
	if(b + 8 < B)
	  c8 = __hfma(x4d(b + 8, m, start_h + p, start_w + q), weights[k], c8);
	if(b + 9 < B)
	  c9 = __hfma(x4d(b + 9, m, start_h + p, start_w + q), weights[k], c9);
	if(b + 10 < B)
	  c10 = __hfma(x4d(b + 10, m, start_h + p, start_w + q), weights[k], c10);
	if(b + 11 < B)
	  c11 = __hfma(x4d(b + 11, m, start_h + p, start_w + q), weights[k], c11);
    

      }
    }

    y4d(b, m, 0, tx) = c0;	
    if(b + 1 < B)
      y4d(b + 1, m, 0, tx) = c1;
    if(b + 2 < B)
      y4d(b + 2, m, 0, tx) = c2;
    if(b + 3 < B)
      y4d(b + 3, m, 0, tx) = c3;
    if(b + 4 < B)
      y4d(b + 4, m, 0, tx) = c4;
    if(b + 5 < B)
      y4d(b + 5, m, 0, tx) = c5;
    if(b + 6 < B)
      y4d(b + 6, m, 0, tx) = c6;
    if(b + 7 < B)
      y4d(b + 7, m, 0, tx) = c7;
    if(b + 8 < B)
      y4d(b + 8, m, 0, tx) = c8;
    if(b + 9 < B)
      y4d(b + 9, m, 0, tx) = c9;
    if(b + 10 < B)
      y4d(b + 10, m, 0, tx) = c10;
    if(b + 11 < B)
      y4d(b + 11, m, 0, tx) = c11;
	
  }
	
  #undef y4d 
  #undef x4d
}





__global__ void depthwise_convNew4_half2(__half* const __restrict__ y,
					const __half* const __restrict__ x,
					const __half* const __restrict__ w,
					const int B, const int M,
					const int H, const int W, const int KH,
					const int KW, const int H_out, const int W_out,
					const int H_pad, const int W_pad,
					const int H_stride, const int W_stride)
{

  #define y4d(i3, i2, i1, i0) y[(i3) * (M * H_out * W_out) + (i2) * (H_out * W_out) + (i1) * (W_out) + i0]
  #define x4d(i3, i2, i1, i0) x[(i3) * (M * H * W) + (i2) * (H * W) + (i1) * (W) + i0]

  const int num = 4;

  const int b = num * blockIdx.x;
  const int m = (blockIdx.y * blockDim.x  + threadIdx.x)/ (H_out * W_out);
	
  if(m < M){
    const int tx = (blockIdx.y * blockDim.x  + threadIdx.x) % (H_out * W_out);

    const int start_h = (tx / W_out) * H_stride - H_pad;
    const int start_w = (tx % W_out) * W_stride - W_pad;

    __half2 c0 = __half2half2(0);
    __half2 c1 = __half2half2(0);
 
    const __half* weights = &w[m * KH * KW];

    for (int k = 0; k < KH * KW; k++) {
      int p = k / KW;
      int q = k % KW;
      if (start_h + p > -1 && start_h + p < H &&
	  start_w + q > -1 && start_w + q < W) {

      
	__half2 t1;
	__half2 t2;
	if(b + 3 < B){
	    t1 = __halves2half2(x4d(b + 1, m, start_h + p, start_w + q), x4d(b, m, start_h + p, start_w + q));
	    t2 = __halves2half2(x4d(b + 3, m, start_h + p, start_w + q), x4d(b + 2, m, start_h + p, start_w + q));
	 }
	else if(b + 2 < B){
	  t1 = __halves2half2(x4d(b + 1, m, start_h + p, start_w + q), x4d(b, m, start_h + p, start_w + q));
	  t2 = __halves2half2(0, x4d(b + 2, m, start_h + p, start_w + q));

	}
	else if(b + 1 < B){
	  t1 = __halves2half2(x4d(b + 1, m, start_h + p, start_w + q), x4d(b, m, start_h + p, start_w + q));
	}
	else{
	  t1 = __halves2half2(0, x4d(b, m, start_h + p, start_w + q));

	 }

	
	c0 = __hfma2(t1, __halves2half2(weights[k], weights[k]), c0);
	c1 = __hfma2(t2, __halves2half2(weights[k], weights[k]), c1);
	
      }
    }

    y4d(b, m, 0, tx) = __high2half(c0);	
    if(b + 1 < B)
      y4d(b + 1, m, 0, tx) = __low2half(c0);
    if(b + 2 < B)
      y4d(b + 2, m, 0, tx) = __high2half(c1);
    if(b + 3 < B)
      y4d(b + 3, m, 0, tx) = __low2half(c1);

  }
	
  #undef y4d 
  #undef x4d
}






// Perforated Tensor Conv with 'perforation_rate' parameter
void* tensorConvPerf(void* input_ptr, void* filter_ptr,
		     int vertical_pad, int horizontal_pad,
		     int vertical_stride, int horizontal_stride,
		     int conv_mode, int conv_groups, int row, int col){

  INFO("*** TensorConvolution \n");
  profileEvent("tensorConv");

  Tensor* input = (Tensor*) input_ptr;
  Tensor* filter = (Tensor*) filter_ptr;

  cudnnConvolutionDescriptor_t convDesc;
  cudnnConvolutionFwdAlgo_t convAlgo;
  cudnnConvolutionMode_t mode;
  if(conv_mode == 0)
    mode = CUDNN_CONVOLUTION;
  else if(conv_mode == 1)
    mode = CUDNN_CROSS_CORRELATION;

  // FIXIT: Need to be more aware of the implications of alpha and beta
  float alpha = 1.0f, beta = 0.0f;

  // TODO: Support other cases;
  hostToDeviceCopy(input);
  hostToDeviceCopy(filter);

  INFO("vertical_stride = %lu, horizontal_stride = %lu \n", vertical_stride, horizontal_stride);

  checkCUDNN(cudnnCreateConvolutionDescriptor(&convDesc));

  //FIXME: Current hack to preserve backward compatibilty
  if(conv_groups == 0){
    conv_groups = 1;
  }

  // NOTE: Adding support for grouped convolution
  checkCUDNN(cudnnSetConvolutionGroupCount(convDesc, conv_groups));

  int new_v = vertical_stride + row;
  int new_h = horizontal_stride + col;
  cudnnDataType_t computeType = CUDNN_DATA_FLOAT;
  // FIXIT: Think if upscaling values need to be configurable?
  // IMP-FIXIT: Either make mode configurable OR see if CUDNN_CONVOLUTION MODE should be used?
  checkCUDNN(cudnnSetConvolution2dDescriptor(convDesc,
					     vertical_pad, horizontal_pad, // conv padding
					     new_v, new_h, // conv strides
					     1, 1, // upscaling values
					     mode , // mode is configurable
					     computeType)); // defines compute precision

  int n, c, h, w; // output dimensions
  // Find dimension of convolution output
  checkCUDNN(cudnnGetConvolution2dForwardOutputDim(convDesc,
						   input->tensor_desc,
						   filter->filter_desc,
						   &n, &c, &h, &w));


  DEBUG("**Output Tensor Dims, n = %d, c = %d, h = %d, w = %d \n", n, c, h, w);

  Tensor* output;
  if(input->data_format == CUDNN_TENSOR_NCHW)
    output = (Tensor*) create4DTensor((cudnnDataType_t) input->data_type,
				      CUDNN_TENSOR_NCHW, n, c, h, w);
  else if(input->data_format == CUDNN_TENSOR_NHWC){
    DEBUG("* NHWC Format \n");
    output = (Tensor*) create4DTensor((cudnnDataType_t) input->data_type,
				      CUDNN_TENSOR_NHWC, n, h, w, c);
  }
  else
    ERROR("Unsupported Tensor Type");

  // NOTE: Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE);
  // NOTE: Necessary to insert the above call for every output tensor

  DEBUG("tensor->data_type = %d, tensor->data_format = %d, N = %d, C = %d, H = %d, W = %d \n",
	output->data_type, output->data_format, output->dims.dim_sizes[0], output->dims.dim_sizes[1],
	output->dims.dim_sizes[2], output->dims.dim_sizes[3]);

  if(convDesc == NULL || input->tensor_desc == NULL ||
     filter->filter_desc == NULL || output->tensor_desc == NULL)
    ERROR("NULL descriptor! \n");


  // Debugging info prints
  printTensorDescInfo(input);
  printTensorDescInfo(filter);
  printTensorDescInfo(output);

  // NOTE-FIXIT: function failing for NHWC formats - perhaps some CUDNN support is lacking
  checkCUDNN(cudnnGetConvolutionForwardAlgorithm(cudnnHandle,
						 input->tensor_desc,
						 filter->filter_desc,
						 convDesc,
						 output->tensor_desc,
						 CUDNN_CONVOLUTION_FWD_PREFER_FASTEST,
						 //CUDNN_CONVOLUTION_FWD_NO_WORKSPACE,
						 0,
						 &convAlgo));


  DEBUG("ConvAlgo = %d, FFT = %d, GEMM = %d, WINOGRAD = %d \n", convAlgo,
	CUDNN_CONVOLUTION_FWD_ALGO_FFT, CUDNN_CONVOLUTION_FWD_ALGO_GEMM,
	CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD);


  // FIXIT: Algo shouldn't be hardcoded
  convAlgo = CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM;

  size_t workspace_size;
  checkCUDNN(cudnnGetConvolutionForwardWorkspaceSize(cudnnHandle,
						     input->tensor_desc,
						     filter->filter_desc,
						     convDesc,
						     output->tensor_desc,
						     convAlgo,
						     &workspace_size));

  // Allocating memory for the convolution workspace
  void* workspace;
  checkCudaErrors(cudaMalloc(&workspace, workspace_size));
  DEBUG("workspace size = %d \n", workspace_size);


  checkCUDNN(cudnnConvolutionForward(cudnnHandle, &alpha, input->tensor_desc,
				     input->gpu_data, filter->filter_desc, filter->gpu_data,
				     convDesc, convAlgo, workspace, workspace_size,
				     &beta, output->tensor_desc, output->gpu_data));


  int old_w = w;
  int old_h = h;
  h = (2 * vertical_pad + input->dims.dim_sizes[2] - filter->dims.dim_sizes[2]) / vertical_stride + 1;
  w = (2 * horizontal_pad + input->dims.dim_sizes[3] - filter->dims.dim_sizes[3]) / horizontal_stride + 1;

  Tensor* new_output;
  if(input->data_format == CUDNN_TENSOR_NCHW)
    new_output = (Tensor*) create4DTensor((cudnnDataType_t) float_type, //input->data_type,
					  CUDNN_TENSOR_NCHW, n, c, h, w);
  else if(input->data_format == CUDNN_TENSOR_NHWC){
    DEBUG("* NHWC Format \n");
    new_output = (Tensor*) create4DTensor((cudnnDataType_t) input->data_type,
					  CUDNN_TENSOR_NHWC, n, h, w, c);
  }
  else
    ERROR("Unsupported Tensor Type");


  int numBlocks = (n * c * h * w  + 127) / 128;
  if(vertical_stride == 0 && row == 0)
    return output;

  if(vertical_stride == 1 && row == 1){
    interpolateRow<<<numBlocks,128>>>(n * c * h * w, old_h, n, c, h, w,
				      (float *)output->gpu_data, (float *)new_output->gpu_data);
  }
  else if(horizontal_stride == 1 && col == 1){
    interpolateCol<<<numBlocks,128>>>(n * c * h * w, old_w, n, c, h, w,
				      (float *)output->gpu_data, (float *)new_output->gpu_data);
  }
  else if (col > 0){
    interpolateXCol<<<numBlocks,128>>>(n * c * h * w, old_w, n, c, h, w,
				       (float *)output->gpu_data, (float *)new_output->gpu_data, col + 1);
  }
  else{
    interpolateXRow<<<numBlocks,128>>>(n * c * h * w, old_h, n, c, h, w,
				       (float *)output->gpu_data, (float *)new_output->gpu_data, row + 1);
  }


  cudaDeviceSynchronize();

  profileEvent("tensorConv_end", true);


  changeTensorPlacement(new_output, DEVICE);
  return new_output;

}








}

