/* This file includes the API implementation of the HPVM tensor runtime built on cublas, cudnn
**
**  Author: Hashim Sharif
**  Email: hsharif3@illinois.edu
*/

#include <stdio.h>
#include <stdarg.h>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <ctime>
#include <cfloat>
#include <algorithm>
#include <chrono>
#include <iomanip>
#include <iostream>
#include <map>
#include <memory>
#include <random>
#include <sstream>
#include <string>
#include <vector>

#include <cuda_runtime.h>
#include <device_launch_parameters.h>

#include <cublas_v2.h>
#include <cudnn.h>
#include <cublas_api.h>
#include <cuda_fp16.h>
#include <driver_types.h>


// Tensor runtime header files
#include "tensor_runtime.h"
#include "tensor_utils.h"
#include "debug.h"
#include "profiling.h"
#include "fp16_conversion.h"
#include "global_data.h"
#include "error.h"
#include "tensor.h"
#include "op_overheads.h"
#include "half_precision_api.h"
#include "hpvm-rt-controller.h"
#include "approxhpvm_runtime_utils.h" 
#include "approx_api.h"


// Image tensor runtime implementation
#include "img_tensor_runtime.cu"

//** Potential Improvements:
//   1) Add support for dataypes beyond floats and half 
//   2) Support for more CUDNN operations



void llvm_hpvm_initTensorRt(int gpuid){

  if(!runtime_initialized){
    
    printf("INITIALIZING GPU %d \n", gpuid);
    // NOTE: Setting the target GPU. Can we use multiple GPUs?
    checkCudaErrors(cudaSetDevice(gpuid));
    // Initializing cuDNN and cuBlas handles
    checkCudaErrors(cublasCreate(&cublasHandle));
    checkCUDNN(cudnnCreate(&cudnnHandle));


#ifdef PROMISE_TUNER_ENABLED
    //    readOpenTunerFlags("opentuner_flags");
    readOpenTunerFlags("promise_flags");
#endif


#ifdef ERROR_INJECTION_ENABLED
    readOpenTunerFlags("opentuner_flags");
#endif

    
    runtime_initialized = true;
  }
  
}


void llvm_hpvm_cleanupTensorRt(){
  DEBUG("\**** llvm_hpvm_cleanupTensorRt ***\n");
  dumpAccuracyNorms();
}


void llvm_hpvm_initApproxhpvmRt(int gpuid){
  llvm_hpvm_initTensorRt(gpuid);
  approxhpvm_runtime_mode = true;
}

void llvm_hpvm_cleanupApproxhpvmRt(){

}



void dumpAccuracyNorms(){

  #ifdef ERROR_INJECTION_ENABLED
  
  
  #endif

  dump_result("accuracy_summary");

}


// Returns the number of GPUs active on the platform
int getGPUCount(){
  int num_gpus;
  checkCudaErrors(cudaGetDeviceCount(&num_gpus));
  return num_gpus;
}



void clearTensorMap(){

  tensors_ptr.clear();
  host_ptr.clear();
  obj_ptr.clear();
}


void startMemTracking(){

  tensors_ptr.clear();
  host_ptr.clear();
  obj_ptr.clear();

  tracked_tensors.clear();
}


void freeOutputTensors(){

  DEBUG("**** Freeing Ouput Tensors *** \n");
  for(int i = 0; i < tensors_ptr.size(); i++){
    cudaFree(tensors_ptr[i]);
    tensors_ptr[i] = NULL;
  }

  for(int i = 0; i < host_ptr.size(); i++){
    free(host_ptr[i]);
    host_ptr[i] = NULL;
  }
  
  for(int i = 0; i < obj_ptr.size(); i++){
    free(obj_ptr[i]);
    obj_ptr[i] = NULL;
  }
}



void clearOpCounter(){
  total_ops = 0;
  op_counter = 0;
  op_accuracies.clear();
}



void freeBatchMemory(){
  // Free allocated memory for the current mini-batch
  freeOutputTensors();
  // Reinitialize couter for OpenTuner flags - next mini-batch of execution
  op_counter = 0;
  // Clearing profiling data map
  func_counters.clear();
}




// FIXIT: Fix any assumptions on the NCHW format
// TODO: benchmark split performance and check if it is prohibitively high?
void** tensorSplit(void* tensor_ptr, int num_splits, int split_dim){

  INFO("*** TensorSplit \n");  
  profileEvent("tensorSplit");

  Tensor* tensor = (Tensor*) tensor_ptr;
  
  deviceToHostCopy(tensor); // Splitting done on the host

  Tensor** splits = (Tensor**) malloc(sizeof(Tensor*) * num_splits);
  size_t* dim_sizes = (size_t*) malloc(sizeof(size_t) * tensor->dims.num_dims);
  for(unsigned int i = 0; i < tensor->dims.num_dims; i++){
    dim_sizes[i] = tensor->dims.dim_sizes[i];
  }

  
  dim_sizes[split_dim] = tensor->dims.dim_sizes[split_dim] / num_splits;
  if(dim_sizes[split_dim] < 1)
    ERROR("Split Dimension < 1 after splitting");

  size_t copy_size = getTypeSize(tensor->data_type);
  for(unsigned int i = split_dim; i < tensor->dims.num_dims; i++){
    copy_size = copy_size * dim_sizes[i];
  }
  
  for(unsigned int i = 0; i < num_splits; i++){
    // FIXIT: Don't be specific to 4D tensors
    // NOTE: Using same data format (NHWC/NCHW) for the split tensors
    INFO("dim_sizes[0] = %d, dim_sizes[1] = %d, dim_sizes[2] = %d, dim_sizes[3] = %d \n",
	 dim_sizes[0], dim_sizes[1], dim_sizes[2], dim_sizes[3]);

    Tensor* split = (Tensor*) create4DTensor(tensor->data_type, tensor->data_format,
					  dim_sizes[0], dim_sizes[1], dim_sizes[2], dim_sizes[3]);
    
    size_t copy_start = i * copy_size;
    size_t copy_stride = num_splits * copy_size;
    INFO("copy_size = %d, copy_start = %d, copy_stride = %d, tensor->size_in_bytes = %d \n",
	 copy_size, copy_start, copy_stride, tensor->size_in_bytes);

    int index = 0;
    while(copy_start + copy_size <= tensor->size_in_bytes){
      memcpy(((char*) split->host_data + (index * copy_size)),
	     ((char*)tensor->host_data + copy_start),
	     copy_size);
      copy_start += copy_stride;
      index++;
    }
   	
    splits[i] = split;     
  }

  profileEvent("tensorSplit_end", true);

  return (void**) splits;
}


void* tensorConcat(void** tensors_ptr, int num_splits, int split_dim){

  INFO("*** TensorConcat \n");  
  profileEvent("tensorConcat");

  Tensor** tensors = (Tensor**) tensors_ptr;

  for(int i = 0; i < num_splits; i++){
    deviceToHostCopy(tensors[i]); // Concatenation done on the host
  }
  
  // The no of dimensions of concatenated tensor are the same
  size_t* dim_sizes = (size_t*) malloc(sizeof(size_t) * tensors[0]->dims.num_dims);
  for(unsigned int i = 0; i < tensors[0]->dims.num_dims; i++){
    dim_sizes[i] = tensors[0]->dims.dim_sizes[i];
  }
  
  size_t copy_size = getTypeSize(tensors[0]->data_type);
  for(unsigned int i = split_dim; i < tensors[0]->dims.num_dims; i++){
    copy_size = copy_size * dim_sizes[i];
  }

  dim_sizes[split_dim] = dim_sizes[split_dim] * num_splits;
  if(dim_sizes[split_dim] < 1)
    ERROR("Split Dimension < 1 after concat");

  Tensor* output = (Tensor*) create4DTensor(tensors[0]->data_type, tensors[0]->data_format,
					 dim_sizes[0], dim_sizes[1], dim_sizes[2], dim_sizes[3]);

  INFO("dim_sizes[0] = %d, dim_sizes[1] = %d, dim_sizes[2] = %d, dim_sizes[3] = %d \n",
       dim_sizes[0], dim_sizes[1], dim_sizes[2], dim_sizes[3]);


  int num_copies = 1;
  for(unsigned int i = 0; i < split_dim; i++){
    num_copies = num_copies * dim_sizes[i];
  }
  
  size_t copy_stride = num_splits * copy_size;
  INFO("copy_size = %d, num_copies = %d, copy_stride = %d, output->size_in_bytes = %d \n",
       copy_size, num_copies, copy_stride, output->size_in_bytes);

  for(unsigned int i = 0; i < num_copies; i++){
    // FIXIT: Don't be specific to 4D tensors
    size_t copy_start = i * copy_stride;
   
    for(int j = 0; j < num_splits; j++){
      struct Tensor* split = tensors[j];
      memcpy(((char*) output->host_data + copy_start + (j * copy_size)),
	     ((char*) split->host_data + (i * copy_size)),
	     copy_size);   
    }      
  }

  profileEvent("tensorConcat_end", true);

  return output;
}



void* tensorLRN(void* input_ptr, unsigned int LRN_window,
		double LRN_alpha, double LRN_beta, double LRN_k){

  INFO("*** TensorLRN \n");  
  profileEvent("tensorLRN");

  Tensor* input = (Tensor*) input_ptr;

  hostToDeviceCopy(input);

  float alpha = 1.0f, beta = 0.0f;
  cudnnLRNDescriptor_t LRNDesc;
  checkCUDNN(cudnnCreateLRNDescriptor(&LRNDesc));

  INFO("window = %d, LRN_alpha = %f, LRN_beta = %f, LRN_k = %f \n",
       LRN_window, LRN_alpha, LRN_beta, LRN_k);
 
  
  checkCUDNN(cudnnSetLRNDescriptor(LRNDesc, LRN_window, LRN_alpha, LRN_beta, LRN_k));

  size_t* dim_sizes = input->dims.dim_sizes;
  Tensor* output = (Tensor*) create4DTensor((cudnnDataType_t) float_type, 
			  CUDNN_TENSOR_NCHW, dim_sizes[0], dim_sizes[1],
			  dim_sizes[2], dim_sizes[3]);
  // NOTE: Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE); 
  // NOTE: Necessary to insert the above call for every output tensor

  printTensorDescInfo(input);
  printTensorDescInfo(output);
  
  checkCUDNN(cudnnLRNCrossChannelForward(cudnnHandle, LRNDesc, CUDNN_LRN_CROSS_CHANNEL_DIM1,
					 &alpha, input->tensor_desc, input->gpu_data,
					 &beta, output->tensor_desc, output->gpu_data));

  profileEvent("tensorLRN_end", true);
    
  return output;
}


void printTensorDims2(void* tensor_ptr){

  struct Tensor* tensor = (struct Tensor*) tensor_ptr;

  printf("Num_elems = %lu \n", tensor->num_elems);
  for (int i = 0; i < tensor->dims.num_dims; i++){
    printf("dim[%d] = %lu \n", i, tensor->dims.dim_sizes[i]);
  }
}




// FIXIT: tensorAdd currently only works for 4D tensors
void* tensorAdd(void* x_ptr, void* bias_ptr){
  
  Tensor* x = (Tensor*) x_ptr;
  Tensor* bias = (Tensor*) bias_ptr;
  
  INFO("*** TensorAdd \n");  
  profileEvent("Add");
    
  float alpha = 1.0f;
  //float beta = 0.0f;
  hostToDeviceCopy(x);
  hostToDeviceCopy(bias);

  convertToFP32(x);
  convertToFP32(bias);

  
  INFO("x->num_elems = %d \n", x->num_elems);
  INFO("bias->num_elems = %d \n", bias->num_elems);

  if(cudnnHandle == NULL){
    ERROR("cudnnHandle NOT initialized!! \n");    
  }
  
  // FIXIT: routine fails for 3D tensors
  checkCUDNN(cudnnAddTensor(cudnnHandle, &alpha, bias->tensor_desc,
			    bias->gpu_data, &alpha, x->tensor_desc, x->gpu_data));

  profileEvent("Add_end", true);

  #ifdef ERROR_INJECTION_ENABLED  
  if(op_counter >= total_ops){
    ERROR("No accuracy flag found \n");
  }
  
  int op_acc = op_accuracies[op_counter];

  // Forcing 0 error for (Resnet-like) equal dimension adds (Testing-only)
  
  //-- if (bias->dims.dim_sizes[0] > 1)
  //--  op_acc = 0;
  // Skip errorInjection if explicitly requested
  //-- if (skip_tensors.find(op_counter) != skip_tensors.end()){
  //--   op_acc = 0;  
  //  }

  void* error_norms = tensorAddError(x, op_acc);
  add_norms(error_norms, "tensorAdd", op_acc);
  add_bias_overheads(x, op_acc);
  op_counter++;
  
  #endif
  
  
  return x;
}


// FIXIT: Generalize all of the routines for types {half, float, double}
void* tensorConvolution(void* input_ptr, void* filter_ptr,
			int vertical_pad, int horizontal_pad,
			int vertical_stride, int horizontal_stride,
			int conv_mode, int conv_groups){  
  
  INFO("*** TensorConvolution \n");
  profileEvent("Conv");

  Tensor* input = (Tensor*) input_ptr;
  Tensor* filter = (Tensor*) filter_ptr;
  
  cudnnConvolutionDescriptor_t convDesc;
  cudnnConvolutionFwdAlgo_t convAlgo;
  cudnnConvolutionMode_t mode;
  if(conv_mode == 0)
    mode = CUDNN_CONVOLUTION;
  else if(conv_mode == 1)
    mode = CUDNN_CROSS_CORRELATION;

  mode = CUDNN_CROSS_CORRELATION;
  // FIXIT: Need to be more aware of the implications of alpha and beta
  float alpha = 1.0f, beta = 0.0f;
  
  // TODO: Support other cases;  
  hostToDeviceCopy(input);
  hostToDeviceCopy(filter);

  convertToFP32(input);
  convertToFP32(filter);

  
  INFO("vertical_stride = %lu, horizontal_stride = %lu \n", vertical_stride, horizontal_stride);  

  checkCUDNN(cudnnCreateConvolutionDescriptor(&convDesc));

  //FIXME: Current hack to preserve backward compatibilty
  if(conv_groups == 0){
    conv_groups = 1;
  }
  
  
  
  cudnnDataType_t computeType = CUDNN_DATA_FLOAT;
  // FIXIT: Think if upscaling values need to be configurable?
  // IMP-FIXIT: Either make mode configurable OR see if CUDNN_CONVOLUTION MODE should be used?
  checkCUDNN(cudnnSetConvolution2dDescriptor(convDesc,
					     vertical_pad, horizontal_pad, // conv padding
					     vertical_stride, horizontal_stride, // conv strides
					     1, 1, // upscaling values
					     mode , // mode is configurable
                                             computeType)); // defines compute precision

  // NOTE: Adding support for grouped convolution
  checkCUDNN(cudnnSetConvolutionGroupCount(convDesc, conv_groups));

  int n, c, h, w; // output dimensions  
  // Find dimension of convolution output

  if(input->tensor_desc == NULL || filter->filter_desc == NULL)
    ERROR("Input or Filter descriptor is NULL");
    
  checkCUDNN(cudnnGetConvolution2dForwardOutputDim(convDesc,
						   input->tensor_desc,
						   filter->filter_desc,
						   &n, &c, &h, &w));

    
  DEBUG("**Output Tensor Dims, n = %d, c = %d, h = %d, w = %d \n", n, c, h, w);

  Tensor* output;
  if(input->data_format == CUDNN_TENSOR_NCHW)
    output = (Tensor*) create4DTensor((cudnnDataType_t) float_type,  
			              CUDNN_TENSOR_NCHW, n, c, h, w);
  else if(input->data_format == CUDNN_TENSOR_NHWC){
    DEBUG("* NHWC Format \n");
    output = (Tensor*) create4DTensor((cudnnDataType_t) float_type, 
			              CUDNN_TENSOR_NHWC, n, h, w, c);
  }
  else
    ERROR("Unsupported Tensor Type");

  // NOTE: Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE); 
  // NOTE: Necessary to insert the above call for every output tensor
    
  DEBUG("tensor->data_type = %d, tensor->data_format = %d, N = %d, C = %d, H = %d, W = %d \n",
	output->data_type, output->data_format, output->dims.dim_sizes[0],
	output->dims.dim_sizes[1],
	output->dims.dim_sizes[2], output->dims.dim_sizes[3]);

  if(convDesc == NULL || input->tensor_desc == NULL ||
     filter->filter_desc == NULL || output->tensor_desc == NULL)
    ERROR("NULL descriptor! \n");


  // Debugging info prints
  printTensorDescInfo(input);
  printTensorDescInfo(filter);
  printTensorDescInfo(output);

  // NOTE-FIXIT: function failing for NHWC formats - perhaps some CUDNN support is lacking
  checkCUDNN(cudnnGetConvolutionForwardAlgorithm(cudnnHandle,
						 input->tensor_desc,
						 filter->filter_desc,
						 convDesc,
						 output->tensor_desc,
						 CUDNN_CONVOLUTION_FWD_PREFER_FASTEST,	 
						 //CUDNN_CONVOLUTION_FWD_NO_WORKSPACE,
						 0,
						 &convAlgo));

  
  DEBUG("ConvAlgo = %d, FFT = %d, GEMM = %d, WINOGRAD = %d \n", convAlgo,
	 CUDNN_CONVOLUTION_FWD_ALGO_FFT, CUDNN_CONVOLUTION_FWD_ALGO_GEMM,
	 CUDNN_CONVOLUTION_FWD_ALGO_WINOGRAD);
	 

  // FIXIT: Algo shouldn't be hardcoded
  //convAlgo = CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_PRECOMP_GEMM;
  convAlgo = CUDNN_CONVOLUTION_FWD_ALGO_IMPLICIT_GEMM;

  size_t workspace_size;
  checkCUDNN(cudnnGetConvolutionForwardWorkspaceSize(cudnnHandle,
						     input->tensor_desc,
						     filter->filter_desc,
						     convDesc,
						     output->tensor_desc,
						     convAlgo,
						     &workspace_size));

  // Allocating memory for the convolution workspace
  void* workspace;
  checkCudaErrors(cudaMalloc(&workspace, workspace_size)); 
  DEBUG("workspace size = %d \n", workspace_size);


  checkCUDNN(cudnnConvolutionForward(cudnnHandle, &alpha, input->tensor_desc,
				     input->gpu_data, filter->filter_desc, filter->gpu_data,
				     convDesc, convAlgo, workspace, workspace_size,
				     &beta, output->tensor_desc, output->gpu_data));
		       
  profileEvent("Conv_end", true);


  #ifdef ERROR_INJECTION_ENABLED

  if(op_counter >= total_ops){
    ERROR("No accuracy flag found \n");
  }
  
  int op_acc = op_accuracies[op_counter];

  // Ignore Error Injection for Depthwise Convolution  
  /*if (conv_groups > 1){
    op_acc = 0;
  }
  */

  
  void* error_norms = tensorAddError(output, op_acc);
  add_norms(error_norms, "tensorConv", op_acc);
  add_conv_overheads(input, filter, vertical_stride, horizontal_stride, op_acc);

  op_counter++;
  
  #endif
  
  
  return output;
}



// NOTE: Supports Max and Avg Pooling
void* tensorPooling(void* input_ptr,
		    int poolFunction,
		    int window_height, int window_width,
		    int vertical_pad, int horizontal_pad,
		    int vertical_stride, int horizontal_stride){

  INFO("*** TensorPooling \n");
  profileEvent("Pool");

  Tensor* input = (Tensor*) input_ptr;

  cudnnPoolingDescriptor_t poolDesc;
  // FIXIT: Need to be more aware of the implications of alpha and beta
  float alpha = 1.0f, beta = 0.0f;

  hostToDeviceCopy(input);

  convertToFP32(input);

  
  checkCUDNN(cudnnCreatePoolingDescriptor(&poolDesc));            

  int n = input->dims.dim_sizes[0];
  int c = input->dims.dim_sizes[1];
  int h = (input->dims.dim_sizes[2] + (2 * vertical_pad) - window_height) / vertical_stride;
  h = h + 1;
  int w = (input->dims.dim_sizes[3] + (2 * horizontal_pad) - window_width) / horizontal_stride;
  w = w + 1;

  DEBUG("n = %d, c = %d, h = %d, w = %d \n", n, c, h, w);
  
  // FIXIT: Don't be specific to floats
  Tensor* output = (Tensor*) create4DTensor(CUDNN_DATA_FLOAT, CUDNN_TENSOR_NCHW, n, c, h, w);
  // Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE); 

  // FIXIT: Fix being specific to CUDNN_DATA_FLOAT and NCHW format
  // FIXIT: Is this setTensor even needed?
  checkCUDNN(cudnnSetTensor4dDescriptor(output->tensor_desc,
					CUDNN_TENSOR_NCHW,
					CUDNN_DATA_FLOAT,
					n, c,
					h, w));


  cudnnPoolingMode_t pool_mode;
  if(poolFunction == 0)
    pool_mode = CUDNN_POOLING_MAX;
  else if(poolFunction == 1)
    pool_mode = CUDNN_POOLING_AVERAGE_COUNT_EXCLUDE_PADDING;

  
  
  // FIXIT: Make the pool function (max, min, avg) configurable
  checkCUDNN(cudnnSetPooling2dDescriptor(poolDesc,
					 //CUDNN_POOLING_MAX,
					 pool_mode,
					 CUDNN_PROPAGATE_NAN,
					 window_height, window_width,
					 vertical_pad, horizontal_pad,
					 vertical_stride, horizontal_stride));
     
  checkCUDNN(cudnnPoolingForward(cudnnHandle, poolDesc, &alpha, input->tensor_desc,
				 input->gpu_data, &beta, output->tensor_desc, output->gpu_data));

  profileEvent("Pool_end", true);


  #ifdef ERROR_INJECTION_ENABLED

  if(op_counter >= total_ops){
    ERROR("No accuracy flag found \n");
  }
  
  int op_acc = op_accuracies[op_counter];
  void* error_norms = tensorAddError(output, op_acc);
  add_norms(error_norms, "tensorPooling", op_acc);
  add_pool_overheads(input, window_height, vertical_stride, op_acc);

  op_counter++;
  
  #endif

  
  return output;
}




void* tensorGemmCPU(void* lhs_ptr, void* rhs_ptr){

  INFO("*** TensorGemmCPU \n");

  Tensor* lhs = (Tensor*) lhs_ptr;
  Tensor* rhs = (Tensor*) rhs_ptr;
  
  // The operation is done on the CPU
  deviceToHostCopy(lhs);
  deviceToHostCopy(rhs);

  if(lhs->data_type != CUDNN_DATA_FLOAT){
    ERROR("Currently only Floating point is supported ");
  }
  
  profileEvent("tensorGemmCPU");
  
  INFO("rhs->dims.num_dims = %d \n", rhs->dims.num_dims);
  INFO("lhs->dims.num_dims = %d \n", lhs->dims.num_dims);

  // FIXIT: Need to be more aware of the implications of alpha and beta
  //float alpha = 1.0f;
  // float beta = 0.0f;
  // 'm' holds the batch dimension - assuming NCHW format Tensors
  int m = lhs->dims.dim_sizes[0];
  // The rhs must be a 2D tensor
  int n = rhs->dims.dim_sizes[rhs->dims.num_dims-1]; // output neurons
  int k = 1;
  // Flattening the dimensions after the batch dimension
  // NOTE: Allowing any number of dimensions > 2 for lhs
  for (int j = 1 ; j < lhs->dims.num_dims; j++){
    k = k * lhs->dims.dim_sizes[j]; // input neurons
  }

  int rhs_k = rhs->dims.dim_sizes[rhs->dims.num_dims-2];
  // Dimension-note: Check if k is same across the two tensors
  INFO("m = %d, n = %d, k = %d \n", m, n, k);
  if(rhs_k != k){
    ERROR("rhs=%d and lhs=%d columns/rows don't match", rhs_k, k);
  }

  // NOTE: Creating a 4D tensor to be compatible with later called cuDNN routines
  Tensor* output = (Tensor*) create4DTensor(CUDNN_DATA_FLOAT, CUDNN_TENSOR_NCHW, m, n, 1, 1);
  // Changing output tensor placement from host to device
  changeTensorPlacement(output, HOST); 

  float* lhs_arr = (float*) lhs->host_data;
  float* rhs_arr = (float*) rhs->host_data;
  float* output_arr = (float*) output->host_data;
  
  for(int i = 0; i < m; i++){
    for(int j = 0; j < n; j++){
      float sum = 0.0;
      for(int l = 0; l < k; l++){
	float mul = lhs_arr[i*k+l] * rhs_arr[l*n+j];
	sum = sum + mul;
      }
      output_arr[i*n+j] = sum;
    }
  }
      
   
  profileEvent("tensorGemmCPU_end", true);
  
  return output;
}



// Reference: https://gist.github.com/peterwittek/6303527
void* tensorGemmGPU(void* lhs_ptr, void* rhs_ptr ){ //, void* result_tensor){

  INFO("*** TensorGemmGPU \n");
  profileEvent("Mul");

  Tensor* lhs = (Tensor*) lhs_ptr;
  Tensor* rhs = (Tensor*) rhs_ptr;


  INFO("rhs->dims.num_dims = %d \n", rhs->dims.num_dims);
  INFO("lhs->dims.num_dims = %d \n", lhs->dims.num_dims);

  // FIXIT: Need to be more aware of the implications of alpha and beta
  float alpha = 1.0f, beta = 0.0f;
  // 'm' holds the batch dimension - assuming NCHW format Tensors
  int m = lhs->dims.dim_sizes[0];
  // The rhs last dimension must contain the neurons
  int n = rhs->dims.dim_sizes[rhs->dims.num_dims-1]; // output neurons
  int k = 1;
  
  // Flattening the dimensions after the batch dimension
  // NOTE: Allowing any number of dimensions > 2 for lhs
  for (int j = 1 ; j < lhs->dims.num_dims; j++){
    k = k * lhs->dims.dim_sizes[j]; // input neurons
  }

  int rhs_k = rhs->dims.dim_sizes[rhs->dims.num_dims-2];
  // Dimension-note: Check if k is same across the two tensors
  INFO("m = %d, n = %d, k = %d \n", m, n, k);
  if(rhs_k != k){
    ERROR("rhs=%d and lhs=%d columns/rows don't match", rhs_k, k);
  }

  Tensor* output = NULL;
  DEBUG("Creating new TENSOR * \n");
  output = (Tensor*) create4DTensor(CUDNN_DATA_FLOAT, CUDNN_TENSOR_NCHW, m, n, 1, 1);

  
  /* else{
    DEBUG("Reusing TENSOR *\n");
    // FIXIT: Add Assertion to check for null pointer and dimension matching
    output = (Tensor*) result_tensor;
    // FIXIT: output value is trashing - Is this deallocated?
    INFO("output->num_elems = %lu \n", output->data_type);
  }
  */
  
  DEBUG("Changing placement *\n");
  // Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE); 

  DEBUG("Changed Placement * \n\n");

  hostToDeviceCopy(lhs);
  hostToDeviceCopy(rhs);

  convertToFP32(lhs);
  convertToFP32(rhs);

  
  DEBUG("CuBlasSgemm *\n");
   
  // INFO: cuBlas uses column-major format
  // INFO: The leading dimension is just the FIRST Dimension
  // IMP: output is N * M in column-major format, M*N in row-major - what cuDNN expects
  checkCudaErrors(cublasSgemm(cublasHandle, CUBLAS_OP_N, CUBLAS_OP_N,
			      n, m, k,
			      &alpha,
			      (float*) rhs->gpu_data, n,
			      (float*) lhs->gpu_data, k,
			      &beta,
			      (float*) output->gpu_data, n));  

  
  profileEvent("Mul_end", true);



  #ifdef ERROR_INJECTION_ENABLED

  if(op_counter >= total_ops){
    ERROR("No accuracy flag found \n");
  }
  
  int op_acc = op_accuracies[op_counter];
  
  void* error_norms = tensorAddError(output, op_acc);
  add_norms(error_norms, "tensorGemm", op_acc);
  add_gemm_overheads(lhs_ptr, rhs_ptr, op_acc);

  op_counter++;
  
  #endif
 
  
  return output;
}







void* tensorGemm(void* lhs_ptr, void* rhs_ptr){

  INFO("*** TensorGemm \n");
  profileEvent("tensorGemm");

  Tensor* lhs = (Tensor*) lhs_ptr;
  Tensor* rhs = (Tensor*) rhs_ptr;
    
  INFO("rhs->dims.num_dims = %d \n", rhs->dims.num_dims);
  INFO("lhs->dims.num_dims = %d \n", lhs->dims.num_dims);

  // FIXIT: Need to be more aware of the implications of alpha and beta
  float alpha = 1.0f, beta = 0.0f;
  // 'm' holds the batch dimension - assuming NCHW format Tensors
  int m = lhs->dims.dim_sizes[0];
  // The rhs last dimension must contain the neurons
  int n = rhs->dims.dim_sizes[rhs->dims.num_dims-1]; // output neurons
  int k = 1;
  // Flattening the dimensions after the batch dimension
  // NOTE: Allowing any number of dimensions > 2 for lhs
  for (int j = 1 ; j < lhs->dims.num_dims; j++){
    k = k * lhs->dims.dim_sizes[j]; // input neurons
  }

  int rhs_k = rhs->dims.dim_sizes[rhs->dims.num_dims-2];
  // Dimension-note: Check if k is same across the two tensors
  INFO("m = %d, n = %d, k = %d \n", m, n, k);
  if(rhs_k != k){
    ERROR("rhs=%d and lhs=%d columns/rows don't match", rhs_k, k);
  }

  // NOTE: Creating a 4D tensor to be compatible with later called cuDNN routines
  Tensor* output = (Tensor*) create4DTensor(CUDNN_DATA_FLOAT, CUDNN_TENSOR_NCHW, m, n, 1, 1);
  // Changing output tensor placement from host to device
  changeTensorPlacement(output, DEVICE); 

  hostToDeviceCopy(lhs);
  hostToDeviceCopy(rhs);

  // NOTE: cuBlas uses column-major format
  // NOTE: The leading dimension is the FIRST Dimension
  // NOTE: The output is N * M in column-major format, M*N in row-major - what cuDNN expects
  checkCudaErrors(cublasSgemm(cublasHandle, CUBLAS_OP_T, CUBLAS_OP_N,
			      n, m, k,
			      &alpha,
			      (float*) rhs->gpu_data, k,
			      (float*) lhs->gpu_data, k,
			      &beta,
			      (float*) output->gpu_data, n));
  
  profileEvent("tensorGemm_end", true);
  
  return output;
}




// FIXIT: Add dimension check assertions throughout the code
void* tensorGemmBias(void* input_ptr, void* bias_ptr){

  INFO("*** TensorGemmBias \n");
  profileEvent("tensorGemmBias");

  Tensor* input = (Tensor*) input_ptr;
  Tensor* bias = (Tensor*) bias_ptr;  

  // NOTE: beta is set to 1 to append to input
  // C = A * B + Beta * C
  float alpha = 1.0f, beta = 1.0f;
  // 'm' holds the batch dimension - assuming NCHW format Tensors
  int m = input->dims.dim_sizes[0];
  // The bias must be a 2D tensor
  int n = bias->dims.dim_sizes[bias->dims.num_dims - 1]; // output neurons

  INFO("m = %d, n = %d \n", m, n);
  
  hostToDeviceCopy(input);
  hostToDeviceCopy(bias);

  struct Tensor* onevec = (Tensor*) create2DTensor(CUDNN_DATA_FLOAT, m, 1);
  fillOnes(onevec);
  hostToDeviceCopy(onevec);
  
  // NOTE: cuBlas uses column-major format
  // NOTE: The leading dimension is just the FIRST Dimension
  checkCudaErrors(cublasSgemm(cublasHandle, CUBLAS_OP_N, CUBLAS_OP_N,
			      n, m, 1,
			      &alpha,
			      (float*) bias->gpu_data, n,
			      (float*) onevec->gpu_data, 1,
 			      &beta,
			      (float*) input->gpu_data, n));

  profileEvent("tensorGemmBias_end", true);
  
  return input;
}


void* tensorRelu(void* input_ptr){

  INFO("*** TensorRelu \n");
  profileEvent("Relu");

  Tensor* input = (Tensor*) input_ptr;
  
  cudnnActivationDescriptor_t reluDesc;
  float alpha = 1.0f, beta = 0.0f;

  hostToDeviceCopy(input);

  convertToFP32(input);
  
  
  checkCUDNN(cudnnCreateActivationDescriptor(&reluDesc));

  checkCUDNN(cudnnSetActivationDescriptor(reluDesc, CUDNN_ACTIVATION_RELU,
					  CUDNN_PROPAGATE_NAN, 0.0));

  checkCUDNN(cudnnActivationForward(cudnnHandle, reluDesc, &alpha,
				    input->tensor_desc, input->gpu_data, &beta,
				    input->tensor_desc, input->gpu_data));

  profileEvent("Relu_end", true);


  #ifdef ERROR_INJECTION_ENABLED
  
  if(op_counter >= total_ops){
    ERROR("No accuracy flag found \n");
  }
  
  int op_acc = op_accuracies[op_counter];
    
  void* error_norms = tensorAddError(input, op_acc);
  add_norms(error_norms, "tensorRelu", op_acc);
  add_relu_overheads(input, op_acc);
  op_counter++;  
  #endif
  

  return input;
}


// Think: Should Softmax be broken into multiple IR operations?
void* tensorSoftmax(void* input_ptr){

  INFO("*** TensorSoftmax \n");
  profileEvent("Softmax");

  Tensor* input = (Tensor*) input_ptr;
  float alpha = 1.0f, beta = 0.0f;

  hostToDeviceCopy(input);
  convertToFP32(input); 
 
    
  // IMP: CUDNN_SOFTMAX_ACCURATE can be replaced with a less acurate CUDNN_SOFTMAX_FAST
  checkCUDNN(cudnnSoftmaxForward(cudnnHandle, CUDNN_SOFTMAX_ACCURATE, CUDNN_SOFTMAX_MODE_CHANNEL,
				 &alpha, input->tensor_desc, input->gpu_data, &beta,
				 input->tensor_desc, input->gpu_data));

  deviceToHostCopy(input);  
  profileEvent("Softmax_end", true);
  
  return input;
}



__global__ void clipValues(float* A, float min, float max, int n){

  int id = blockIdx.x * blockDim.x + threadIdx.x;

  if(id < n){
    A[id] = fmaxf(min, A[id]);
    A[id] = fminf(max, A[id]);
  }
}



void* tensorRelu2(void* input_ptr, float min, float max){

  INFO("*** TensorClippedRelu *** \n");
  profileEvent("Relu");

  cudnnActivationDescriptor_t reluDesc;
  float alpha = 1.0f, beta = 0.0f;
  
  Tensor* input = (Tensor*) input_ptr;

  hostToDeviceCopy(input);

  convertToFP32(input);
  

  checkCUDNN(cudnnCreateActivationDescriptor(&reluDesc));

  checkCUDNN(cudnnSetActivationDescriptor(reluDesc, CUDNN_ACTIVATION_CLIPPED_RELU,
					  CUDNN_PROPAGATE_NAN, max));

  checkCUDNN(cudnnActivationForward(cudnnHandle, reluDesc, &alpha,
				    input->tensor_desc, input->gpu_data, &beta,
				    input->tensor_desc, input->gpu_data));

  
  
  profileEvent("Relu_end", true);


  #ifdef ERROR_INJECTION_ENABLED
  
  if(op_counter >= total_ops){
    ERROR("No accuracy flag found \n");
  }
  
  int op_acc = op_accuracies[op_counter];
  void* error_norms = tensorAddError(input, op_acc);
  add_norms(error_norms, "tensorClippedRelu", op_acc);
  add_relu_overheads(input, op_acc);
  op_counter++;  
  #endif
  

  return input;
}


void* tensorTanh(void* input_ptr){

  INFO("*** TensorTanh \n");
  profileEvent("Tanh");

  Tensor* input = (Tensor*) input_ptr;
  
  cudnnActivationDescriptor_t tanhDesc;
  float alpha = 1.0f, beta = 0.0f;

  hostToDeviceCopy(input);

  convertToFP32(input);

  
  checkCUDNN(cudnnCreateActivationDescriptor(&tanhDesc));

  checkCUDNN(cudnnSetActivationDescriptor(tanhDesc, CUDNN_ACTIVATION_TANH,
					  CUDNN_PROPAGATE_NAN, 0.0));

  checkCUDNN(cudnnActivationForward(cudnnHandle, tanhDesc, &alpha,
				    input->tensor_desc, input->gpu_data, &beta,
				    input->tensor_desc, input->gpu_data));

  profileEvent("Tanh_end", true);


  #ifdef ERROR_INJECTION_ENABLED
  
  if(op_counter >= total_ops){
    ERROR("No accuracy flag found \n");
  }
  
  int op_acc = op_accuracies[op_counter];
  void* error_norms = tensorAddError(input, op_acc);
  add_norms(error_norms, "tensorTanh", op_acc);
  add_relu_overheads(input, op_acc);
  op_counter++;  
  #endif
  

  return input;
}




void* tensorBatchNorm(void* input_ptr, void* gamma_ptr, void* beta_ptr,
		      void* mean_ptr, void* variance_ptr, double epsilon){

  INFO("*** TensorBatchNorm \n");
  profileEvent("BatchNorm");

  Tensor* input = (Tensor*) input_ptr;
  Tensor* gamma = (Tensor*) gamma_ptr;
  Tensor* beta = (Tensor*) beta_ptr;
  Tensor* mean = (Tensor*) mean_ptr;
  Tensor* variance = (Tensor*) variance_ptr;

  if (input == NULL || gamma == NULL || beta == NULL || mean == NULL || variance == NULL){
    ERROR("NULL Input Tensor");
  }
  
  float alpha_val = 1.0f, beta_val = 0.0f;
  hostToDeviceCopy(input);
  hostToDeviceCopy(gamma);
  hostToDeviceCopy(beta);
  hostToDeviceCopy(mean);
  hostToDeviceCopy(variance);

  convertToFP32(input);

 
  
  checkCUDNN(cudnnBatchNormalizationForwardInference(cudnnHandle, CUDNN_BATCHNORM_SPATIAL,
						     &alpha_val, &beta_val,
						     input->tensor_desc, input->gpu_data,
						     input->tensor_desc, input->gpu_data,
						     gamma->tensor_desc, gamma->gpu_data,
						     beta->gpu_data, mean->gpu_data,
						     variance->gpu_data,
						     epsilon));

  profileEvent("BatchNorm_end", true);


  #ifdef ERROR_INJECTION_ENABLED
  
  if(op_counter >= total_ops){
    ERROR("No accuracy flag found \n");
  }
  
  int op_acc = op_accuracies[op_counter];
  //op_acc = 0;  
  void* error_norms = tensorAddError(input, op_acc);
  add_norms(error_norms, "tensorBatchNorm", op_acc);
  add_relu_overheads(input, op_acc);
  op_counter++;  
  #endif
  

  return input;
}




/************* GPU Layer API  *************/

void* ConvLayer_GPU(void* input, 
		    void* filter, 
		    void* bias, 
		    int conv_pad_h, int conv_pad_w, int conv_stride_h, int conv_stride_w,
		    int pool_id, int pool_size,
		    int activation_id, // Relu, Tanh, ClipRelu
		    float out_min, float out_max){ // NOTE: min_val, max_val apply to 'ClippedRelu'

  void* conv_out = tensorConvolution(input, filter,
				     conv_pad_h, conv_pad_w,
				     conv_stride_h, conv_stride_w,
				     1, 0);
  void* conv_add;
  if(bias != NULL){
    conv_add = tensorAdd(conv_out, bias);
  }
  else{
    conv_add = conv_out;
  }

  void* activation_out;  
  switch(activation_id){
  case -1:
    activation_out = conv_add;
    INFO("NO Activation Function \n");
    break;
  case 0:
    activation_out = tensorTanh(conv_add);
    break;
  case 1:
    activation_out = tensorRelu(conv_add);
    break;
  case 2:
    activation_out = tensorRelu2(conv_add, out_min, out_max);
    break;
  default:
    ERROR("Activation id %d NOT supported \n", activation_out);
    break;
  }


  void* pool_out = activation_out;
  // NOTE: Skip pooling on negative pool sizes
  if(pool_size > 0){
    //FIXME: Currently only using MaxPooling
    pool_out = tensorPooling(activation_out, 0, pool_size, pool_size, 0, 0, pool_size, pool_size);
  }
  else{
    pool_out = activation_out;
  }

  return pool_out;
}


void* FCLayer_GPU(void* input, 
		  void* weights, 
		  void* bias, 
		  int activation_id,
		  float out_min, float out_max){ // NOTE: min_val, max_val apply to 'ClippedRelu'

  void* gemm_out = tensorGemmGPU(input, weights);

  void* gemmbias_out;
  if(bias != NULL){
    gemmbias_out = tensorAdd(gemm_out, bias);
  }
  else{
    gemmbias_out = gemm_out;
  }
 
  void* activation_out;
  switch(activation_id){

  case -1:
    activation_out = gemmbias_out;
    INFO("No Activation Function \n");
    break;
  case 0:
    activation_out = tensorTanh(gemmbias_out);
    break;
  case 1:
    activation_out = tensorRelu(gemmbias_out);
    break;
  case 2:
    activation_out = tensorRelu2(gemmbias_out, out_min, out_max);
    break;
  default:
    ERROR("Activation id %d NOT supported \n", activation_out);
    break;
  }
   
  return activation_out;
}


/*********** PROMISE API **************/

/*
void* ConvLayer_PROMISE(void* input, float i_min, float i_max,
			void* filter, float w_min, float w_max,
			void* bias, float b_min, float b_max,
			int conv_pad_h, int conv_pad_w, int conv_stride_h, int conv_stride_w,
			int pool_id, int pool_size,
			int activation_id, // Relu, Tanh, ClipRelu
			float out_min, float out_max, int swing){ 

  
  #ifdef PROMISE_TUNER_ENABLED

  // NOTE: Skip reading file-based error levels for ApproxHPVM wrapper runtime
  if(!approxhpvm_runtime_mode){
  
    if(op_counter >= total_ops){
      ERROR("No accuracy flag found \n");
    }
  
    swing = op_accuracies[op_counter];
    op_counter++;
  }
  
  #endif  

  
  if (swing < 0 || swing > 20){
    ERROR("Incorrect swing value");
  }

  

  if(swing < 8){
    input = quantizeTensorPromise(input, i_min, i_max);
    filter = quantizeTensorPromise(filter, w_min, w_max);
    if(bias != NULL)
      bias = quantizeTensorPromise(bias, b_min, b_max);
    // aRead error
    
    input = addPromiseError(input, swing);
  }

  
  void* conv_out;
  if(swing == 8 || (swing >= 12 && swing <= 15) ){
    //conv_out = tensorConvPerf(input, filter, conv_pad_h, conv_pad_w,
    //		              conv_stride_h, conv_stride_w, 1, 1, 1, 0);

    int rows = 2;
    switch(swing){

    case 12: rows = 5; break;
    case 13: rows = 4; break;
    case 14: rows = 3; break;
    case 15: rows = 2; break;    
		   
    default: rows = 2; break;
    }
    
    conv_out = tensorConvPerf2(input, filter, conv_pad_h, conv_pad_w,
    		              conv_stride_h, conv_stride_w, 1, 1, rows, 0);

    /*void* gold = tensorConvolution(input, filter,
				   conv_pad_h, conv_pad_w,
				   conv_stride_h, conv_stride_w,
				   1, 0);

    Norm_t* norms = calculateNormsTreeReduction((struct Tensor*) conv_out, (struct Tensor*) gold);

    DEBUG("\n-------- l2_norm = %f \n", norms->l2_norm); 
    */



  /* -----
  }
  else if(swing == 9 || (swing >= 16 && swing <= 19) ){
    //conv_out = tensorConvPerf(input, filter, conv_pad_h, conv_pad_w,
    //		              conv_stride_h, conv_stride_w, 1, 1, 0, 1);


    int cols = 2;
    switch(swing){

    case 16: cols = 5; break;
    case 17: cols = 4; break;
    case 18: cols = 3; break;
    case 19: cols = 2; break;    
		   
    default: cols = 2; break;
    }

    
    conv_out = tensorConvPerf2(input, filter, conv_pad_h, conv_pad_w,
    		              conv_stride_h, conv_stride_w, 1, 1, 0, cols);


    /*void* gold = tensorConvolution(input, filter,
				   conv_pad_h, conv_pad_w,
				   conv_stride_h, conv_stride_w,
				   1, 0);

    Norm_t* norms = calculateNormsTreeReduction((struct Tensor*)conv_out, (struct Tensor*) gold);

    DEBUG("\n-------- l2_norm = %f \n", norms->l2_norm); 
    */

  /*------
  }
  else if(swing == 10){  
    conv_out = tensorHalfConvolution(input, filter,
				     conv_pad_h, conv_pad_w,
				     conv_stride_h, conv_stride_w,
				     1, 0);
  }
  else{
    conv_out = tensorConvolution(input, filter,
				 conv_pad_h, conv_pad_w,
				 conv_stride_h, conv_stride_w,
				 1, 0);
  }
  
  void* conv_add;
  if(bias != NULL){
    if(swing >= 8){  
      conv_add = tensorHalfAdd(conv_out, bias);
    }
    else{
      conv_add = tensorAdd(conv_out, bias);
    }
  }
  else{
    conv_add = conv_out;
  }

  void* pool_out;
  // NOTE: Skip pooling on negative pool sizes
  if(pool_size > 0){
    //FIXME: Currently only using MaxPooling
    pool_out = tensorHalfPooling(conv_add, 0, pool_size, pool_size, 0, 0, pool_size, pool_size);
  }
  else{
    pool_out = conv_add;
  }
  
  void* activation_out;  
  switch(activation_id){
  case -1:
    activation_out = pool_out;
    INFO("NO Activation Function \n");
    break;
  case 0:
    activation_out = tensorHalfTanh(pool_out);
    break;
  case 1:
    activation_out = tensorHalfRelu(pool_out);
    break;
  case 2:
    activation_out = tensorHalfRelu2(pool_out, out_min, out_max);
    break;
  default:
    ERROR("Activation id %d NOT supported \n", activation_out);
    break;
  }


  if(swing < 8 && activation_id != -1){
    activation_out = quantizeTensorPromise(activation_out, out_min, out_max);
  }
  
  return activation_out;
}


void* FCLayer_PROMISE(void* input, float i_min, float i_max,
		      void* weights, float w_min, float w_max,
		      void* bias, float b_min, float b_max,
		      int activation_id,
		      float out_min, float out_max, int swing){ //NOTE: min_val, max_val apply to 'ClippedRelu'


  
  #ifdef PROMISE_TUNER_ENABLED

  // NOTE: Skip reading file-based error levels for ApproxHPVM wrapper runtime
  if(!approxhpvm_runtime_mode){

    if(op_counter >= total_ops){
      ERROR("No accuracy flag found \n");
    }
  
    swing = op_accuracies[op_counter];
    op_counter++;
  }
  
  #endif
 
  
  if (swing < 0 || swing > 20){
    ERROR("Incorrect swing value");
  }
  
  if(swing < 8){
    input = quantizeTensorPromise(input, i_min, i_max);
    weights = quantizeTensorPromise(weights, w_min, w_max);
    if(bias != NULL)
      bias = quantizeTensorPromise(bias, b_min, b_max);

    // NOTE: Modelling aRead error in PROMISE
    input = addPromiseError(input, swing);
  }


  
  void* gemm_out;
  if(swing >= 8 && swing < 11){
    gemm_out = tensorHalfGemm(input, weights);
  }
  else{
    gemm_out = tensorGemmGPU(input, weights);
  }

  
  void* gemmbias_out;
  if(bias != NULL){
    // Swing 8 corresponds to FP32
    if(swing >= 8 && swing < 20){
      gemmbias_out = tensorHalfAdd(gemm_out, bias);
    }
    else{
      gemmbias_out = tensorAdd(gemm_out, bias);
    }
  }
  else{
    gemmbias_out = gemm_out;
  }
 
  void* activation_out;
  switch(activation_id){

  case -1:
    activation_out = gemmbias_out;
    INFO("No Activation Function \n");
    break;
  case 0:
    activation_out = tensorTanh(gemmbias_out);
    break;
  case 1:
    activation_out = tensorRelu(gemmbias_out);
    break;
  case 2:
    activation_out = tensorRelu2(gemmbias_out, out_min, out_max);
    break;
  default:
    ERROR("Activation id %d NOT supported \n", activation_out);
    break;
  }
  
  
  if(swing < 8 && activation_id != -1){
    activation_out = quantizeTensorPromise(activation_out, out_min, out_max);
  }
  
  return activation_out;
}

*****/



/**** Wrapper Runtime API ***/
  
void* wrapper_ConvLayer(const char* hpvm_node_id,
			void* input, 
		        void* filter, 
		        void* bias, 
		        int conv_pad_h, int conv_pad_w,
		        int conv_stride_h, int conv_stride_w,
		        int pool_id, int pool_size,
		        int activation_id,
		        // NOTE: out_min, out_max are only relevant for ClippedRelu
		        float out_min, float out_max){

  NodeConfiguration *NodeConf = RC->getNodeConfiguration(hpvm_node_id);

  if (NodeConf->isPROMISENodeConfiguration()) {
    DEBUG("PROMISE Configuration for ConvLayer\n");
    // Mapped to PROMISE - get a PROMISE node configuration
    PROMISENodeConfiguration *PROMISEConf = (PROMISENodeConfiguration *)NodeConf;
    std::vector<float> &QRanges = RC->getQuantizationRanges(hpvm_node_id);

    std::vector<std::pair<PROMISENodeConfiguration::APPROX, int> > &approxTuples =
      PROMISEConf->getApproxChoices();

    if (approxTuples.size() == 1) {
      enum PROMISENodeConfiguration::APPROX approx = approxTuples[0].first;
      int param = approxTuples[0].second;
      if (approx == PROMISENodeConfiguration::APPROX::SWING_LEVEL) {
        DEBUG("Approximation choice for ConvLayer: swing level %d\n", param);

        struct Tensor* input_tensor_cast = (struct Tensor*) input;
        struct Tensor* filter_tensor_cast = (struct Tensor*) filter;
        std::pair<double, double> pinfo =
          RC->conv_profile(input_tensor_cast->dims.dim_sizes[0], //n
                           input_tensor_cast->dims.dim_sizes[1], //c
                           input_tensor_cast->dims.dim_sizes[2], //h
                           input_tensor_cast->dims.dim_sizes[3], //w
                           filter_tensor_cast->dims.dim_sizes[0], //c_out
                           filter_tensor_cast->dims.dim_sizes[1], //c_in
                           filter_tensor_cast->dims.dim_sizes[2], //k_h
                           filter_tensor_cast->dims.dim_sizes[3], //k_w
                           conv_stride_h, //s_h
                           conv_stride_w, //s_w
                           param, //voltage_swing
                           filter_tensor_cast->dims.dim_sizes[2] *
                             filter_tensor_cast->dims.dim_sizes[3] /*patch_factor: k_h*k_w*/);
        RC->addToCurrentIterationComputeTime("ConvLayer_PROMISE", pinfo.first);
        RC->addToCurrentIterationComputeEnergy("ConvLayer_PROMISE", pinfo.second);
        void* t_out;
        t_out = PROMISE_Conv(input, QRanges[0], QRanges[1],
                            filter, QRanges[2], QRanges[3],
                            bias, QRanges[4], QRanges[5],
                            conv_pad_h, conv_pad_w,
                            conv_stride_h, conv_stride_w,
                            pool_id, pool_size,
                            activation_id,
                            QRanges[6], QRanges[7], param);

        return t_out;
      } else {
        CUSTOM_ASSERT(false && "Unknown approximation type");
        ERROR("Unknown approximation type");
        abort();
      }
      // TODO additional approx methods implemented here

    } else if (approxTuples.size() == 2) {
      ERROR("Currently unsupported case");
      abort();
    } else {
      ERROR("Unsupported case");
      abort();
    }
  }
  else
  if (NodeConf->isGPUNodeConfiguration()) {
    DEBUG("GPU Configuration for ConvLayer\n");
    // Mapped to GPU - get a GPU node configuration
    GPUNodeConfiguration *GPUConf = (GPUNodeConfiguration *)NodeConf;

    std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                            std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                   int> > > > &ApproxChoices =
      GPUConf->getApproxChoices();

    // Check for convolution as first operation
    CUSTOM_ASSERT((ApproxChoices.size() >= 1) &&
           (ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::CONV) &&
           "Incorrect number/type of operations in provided Conv layer configuration");

    void* conv_out = handleTensorConvApproximationTuples(ApproxChoices[0].second,
                       input, filter, conv_pad_h, conv_pad_w,
                       conv_stride_h, conv_stride_w);
    void* add_out;
    if (bias != NULL) {
      // Check for add as second operation
      CUSTOM_ASSERT((ApproxChoices.size() >= 2) &&
             (ApproxChoices[1].first == GPUNodeConfiguration::TENSOR_OP::ADD) &&
             "Incorrect number/type of operations in provided Conv layer configuration");
      add_out = handleTensorAddApproximationTuples(ApproxChoices[1].second,
                                                   conv_out, bias);
    } else {
      add_out = conv_out;
    }

    void* activation_out;
    switch (activation_id) {
      case -1:
        { // No activation
          INFO("No activation Function\n");
          activation_out = add_out;
        }
        break;
      case 0:
        { // TanH activation
          CUSTOM_ASSERT((ApproxChoices.size() >= 3) &&
                 (ApproxChoices[2].first == GPUNodeConfiguration::TENSOR_OP::TANH) &&
                 "Incorrect number/type of operations in provided Conv layer configuration");
          activation_out = handleTensorTanhApproximationTuples(ApproxChoices[2].second,
                                                               add_out);
        }
        break;
      case 1:
        { // ReLU activation
          CUSTOM_ASSERT((ApproxChoices.size() >= 3) &&
                 (ApproxChoices[2].first == GPUNodeConfiguration::TENSOR_OP::RELU) &&
                 "Incorrect number/type of operations in provided Conv layer configuration");
          activation_out = handleTensorReluApproximationTuples(ApproxChoices[2].second,
                                                               add_out);
        }
        break;
      case 2:
        { // Clipped ReLU activation
          CUSTOM_ASSERT((ApproxChoices.size() >= 3) &&
                 (ApproxChoices[2].first == GPUNodeConfiguration::TENSOR_OP::CLIPPED_RELU) &&
                 "Incorrect number/type of operations in provided Conv layer configuration");
          activation_out =
            handleTensorClippedReluApproximationTuples(ApproxChoices[2].second,
                                                       add_out, out_min, out_max);
        }
        break;
      default:
        {
          ERROR("Activation id %d NOT supported \n", activation_id);
        }
        break;
    }

    void* pool_out;

    if (pool_size > 0) {
      switch (pool_id) {
        case 0:
          {
            // If we remove the asserts, we can have all cases handled by a single call
            CUSTOM_ASSERT((ApproxChoices.back().first == GPUNodeConfiguration::TENSOR_OP::POOL_MAX) &&
                  "Expected POOL_MAX in provided Conv layer configuration");
            pool_out =
              handleTensorPoolingApproximationTuples(ApproxChoices.back().second,
                                                     activation_out, pool_id,
                                                     pool_size, pool_size, 0, 0,
                                                     pool_size, pool_size);
          }
          break;
        case 1:
          {
            CUSTOM_ASSERT((ApproxChoices.back().first == GPUNodeConfiguration::TENSOR_OP::POOL_MEAN) &&
                  "Expected POOL_MEAN in provided Conv layer configuration");
            pool_out =
              handleTensorPoolingApproximationTuples(ApproxChoices.back().second,
                                                     activation_out, pool_id,
                                                     pool_size, pool_size, 0, 0,
                                                     pool_size, pool_size);
          }
          break;
        case 2:
          {
            CUSTOM_ASSERT((ApproxChoices.back().first == GPUNodeConfiguration::TENSOR_OP::POOL_MIN) &&
                  "Expected POOL_MIN in provided Conv layer configuration");
            pool_out =
              handleTensorPoolingApproximationTuples(ApproxChoices.back().second,
                                                     activation_out, pool_id,
                                                     pool_size, pool_size, 0, 0,
                                                     pool_size, pool_size);
          }
          break;
        default:
          {
            ERROR("Pool id %d NOT supported \n", pool_id);
          }
          break;
      }
    } else {
      pool_out = activation_out;
    }
    return pool_out;
  }
  else {
    ERROR("Unsupported Configuration");
    abort();
  }

  return NULL;
}


void* wrapper_FCLayer(const char* hpvm_node_id,
		      void* input, 
		      void* weights, 
		      void* bias, 
		      int activation_id,
		      // NOTE: out_min and out_max are only relevant for ClippedRelu
		      float out_min, float out_max){ 

  NodeConfiguration *NodeConf = RC->getNodeConfiguration(hpvm_node_id);

  if (NodeConf->isPROMISENodeConfiguration()) {
    DEBUG("PROMISE Configuration for FCLayer\n");
    // Mapped to PROMISE - get a PROMISE node configuration
    PROMISENodeConfiguration *PROMISEConf = (PROMISENodeConfiguration *)NodeConf;
    std::vector<float> &QRanges = RC->getQuantizationRanges(hpvm_node_id);

    std::vector<std::pair<PROMISENodeConfiguration::APPROX, int> > &approxTuples =
      PROMISEConf->getApproxChoices();

    if (approxTuples.size() == 1) {
      enum PROMISENodeConfiguration::APPROX approx = approxTuples[0].first;
      int param = approxTuples[0].second;
      if (approx == PROMISENodeConfiguration::APPROX::SWING_LEVEL) {
        DEBUG("Approximation choice for FCLayer: swing level %d\n", param);

        struct Tensor* input_tensor_cast = (struct Tensor*) input;
        struct Tensor* weights_tensor_cast = (struct Tensor*) weights;
        CUSTOM_ASSERT((input_tensor_cast->dims.dim_sizes[1] *
                       input_tensor_cast->dims.dim_sizes[2] *
                       input_tensor_cast->dims.dim_sizes[3] ==
                         weights_tensor_cast->dims.dim_sizes[2]) &&
                      "Dimensions for matrix multiplication do not match.");
        std::pair<double, double> pinfo =
          RC->fc_profile(input_tensor_cast->dims.dim_sizes[0], //num_rows_a,
                         input_tensor_cast->dims.dim_sizes[1] *
                           input_tensor_cast->dims.dim_sizes[2] *
                           input_tensor_cast->dims.dim_sizes[3], //num_cols_a,
                         weights_tensor_cast->dims.dim_sizes[2], //num_rows_b,
                         weights_tensor_cast->dims.dim_sizes[3], //num_cols_b,
                         param, //voltage_swing,
                         1 /*patch_factor*/);
        RC->addToCurrentIterationComputeTime("FCLayer_PROMISE", pinfo.first);
        RC->addToCurrentIterationComputeEnergy("FCLayer_PROMISE", pinfo.second);
        void* t_out;
        t_out = PROMISE_FC(input, QRanges[0], QRanges[1],
                           weights, QRanges[2], QRanges[3],
                           bias, QRanges[4], QRanges[5],
                           activation_id,
                           QRanges[6], QRanges[7], param);
        return t_out;
      } else {
        CUSTOM_ASSERT(false && "Unknown approximation type");
        ERROR("Unknown approximation type");
        abort();
      }
      // TODO additional approx methods implemented here

    } else if (approxTuples.size() == 2) {
      ERROR("Currently unsupported case");
      abort();
    } else {
      ERROR("Unsupported case");
      abort();
    }
  }
  else
  if (NodeConf->isGPUNodeConfiguration()) {
    DEBUG("GPU Configuration for FCLayer\n");
    // Mapped to GPU - get a GPU node configuration
    GPUNodeConfiguration *GPUConf = (GPUNodeConfiguration *)NodeConf;

    std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                            std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                   int> > > > &ApproxChoices =
      GPUConf->getApproxChoices();

    // Approximation choices must be for a FC wrapper operation
    CUSTOM_ASSERT((ApproxChoices.size() == 2 || ApproxChoices.size() == 3) &&
         ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::MUL &&
         ApproxChoices[1].first == GPUNodeConfiguration::TENSOR_OP::ADD &&
         "Invalid configuration generated for FC layer wrapper operation");

    void* gemm_out = handleTensorMulApproximationTuples(ApproxChoices[0].second,
                                                        input, weights);
    void* add_out = handleTensorAddApproximationTuples(ApproxChoices[1].second,
                                                        gemm_out, bias);

    void* activation_out;
    switch (activation_id) {
      case -1:
        { // No activation
          CUSTOM_ASSERT((ApproxChoices.size() == 2) &&
                 "Incorrect number of operations in provided FC layer configuration");
          INFO("No activation Function\n");
          activation_out = add_out;
        }
        break;
      case 0:
        { // TanH activation
          CUSTOM_ASSERT((ApproxChoices.size() == 3) &&
                 (ApproxChoices[2].first == GPUNodeConfiguration::TENSOR_OP::TANH) &&
                 "Incorrect number/type of operations in provided FC layer configuration");
          activation_out = handleTensorTanhApproximationTuples(ApproxChoices[1].second,
                                                               add_out);
        }
        break;
      case 1:
        { // ReLU activation
          CUSTOM_ASSERT((ApproxChoices.size() == 3) &&
                 (ApproxChoices[2].first == GPUNodeConfiguration::TENSOR_OP::RELU) &&
                 "Incorrect number/type of operations in provided FC layer configuration");
          activation_out = handleTensorReluApproximationTuples(ApproxChoices[1].second,
                                                               add_out);
        }
        break;
      case 2:
        { // Clipped ReLU activation
          CUSTOM_ASSERT((ApproxChoices.size() == 3) &&
                 (ApproxChoices[2].first == GPUNodeConfiguration::TENSOR_OP::CLIPPED_RELU) &&
                 "Incorrect number/type of operations in provided FC layer configuration");
          activation_out =
            handleTensorClippedReluApproximationTuples(ApproxChoices[1].second,
                                                       add_out, out_min, out_max);
        }
        break;
      default:
        {
          ERROR("Activation id %d NOT supported \n", activation_id);
        }
        break;
    }
    return activation_out;
  }
  else {
    ERROR("Unsupported Configuration");
    abort();
  }

  return NULL;
}




void* wrapper_tensorRelu(const char* hpvm_node_id, void* input_ptr){
//  return tensorRelu(input_ptr);

  // Only mapped to GPU - get a GPU configuration
  GPUNodeConfiguration *GPUConf =
    (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);

  std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                          std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                 int> > > > &ApproxChoices =
    GPUConf->getApproxChoices();

  // Approximation choices must be for a relu operation
  CUSTOM_ASSERT(ApproxChoices.size() == 1 &&
         ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::RELU &&
         "Invalid configuration generated for tensor relu wrapper operation");

  return handleTensorReluApproximationTuples(ApproxChoices[0].second,
                                             input_ptr);

}

void* wrapper_tensorClippedRelu(const char* hpvm_node_id,
                                void* input_ptr,
                                float out_min, float out_max){
  // Only mapped to GPU - get a GPU configuration
  GPUNodeConfiguration *GPUConf =
    (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);

  std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                          std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                 int> > > > &ApproxChoices =
    GPUConf->getApproxChoices();

  // Approximation choices must be for a relu operation
  CUSTOM_ASSERT(ApproxChoices.size() == 1 &&
         ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::CLIPPED_RELU &&
         "Invalid configuration generated for tensor clipped relu wrapper operation");

  return handleTensorClippedReluApproximationTuples(ApproxChoices[0].second,
                                                    input_ptr, out_min, out_max);

}

void* wrapper_tensorTanh(const char* hpvm_node_id, void* input_ptr){
//  return tensorTanh(input_ptr);

  GPUNodeConfiguration *GPUConf =
    (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);

  std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                          std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                 int> > > > &ApproxChoices =
    GPUConf->getApproxChoices();

  // Approximation choices must be for a tanh operation
  CUSTOM_ASSERT(ApproxChoices.size() == 1 &&
         ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::TANH &&
         "Invalid configuration generated for tensor tanh wrapper operation");

  return handleTensorTanhApproximationTuples(ApproxChoices[0].second,
                                             input_ptr);

}


void* wrapper_tensorBatchNorm(const char* hpvm_node_id,
			      void* input_ptr, void* gamma_ptr, void* beta_ptr,
			      void* mean_ptr, void* variance_ptr, double epsilon){
//  return tensorBatchNorm(input_ptr, gamma_ptr, beta_ptr, mean_ptr, variance_ptr, epsilon);

  // Only mapped to GPU - get a GPU configuration
  GPUNodeConfiguration *GPUConf =
    (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);

  std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                          std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                 int> > > > &ApproxChoices =
    GPUConf->getApproxChoices();

  // Approximation choices must be for a batchnorm operation
  CUSTOM_ASSERT(ApproxChoices.size() == 1 &&
         ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::BATCHNORM &&
         "Invalid configuration generated for tensor batchnorm wrapper operation");

  return handleTensorBatchNormApproximationTuples(ApproxChoices[0].second,
                                                  input_ptr, gamma_ptr, beta_ptr,
                                                  mean_ptr, variance_ptr, epsilon);

}


void* wrapper_tensorAdd(const char* hpvm_node_id, void* input_ptr, void* bias_ptr){
//  return tensorAdd(input_ptr, bias_ptr);

  // Only mapped to GPU - get a GPU configuration
  GPUNodeConfiguration *GPUConf =
    (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);

  std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                          std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                 int> > > > &ApproxChoices =
    GPUConf->getApproxChoices();

  // Approximation choices must be for an add operation
  CUSTOM_ASSERT(ApproxChoices.size() == 1 &&
         ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::ADD &&
         "Invalid configuration generated for tensor add wrapper operation");

  return handleTensorAddApproximationTuples(ApproxChoices[0].second,
                                            input_ptr, bias_ptr);

}


void* wrapper_tensorPooling(const char* hpvm_node_id,
			    void* input_ptr,
			    int poolFunction,
			    int window_height, int window_width,
			    int vertical_pad, int horizontal_pad,
			    int vertical_stride, int horizontal_stride){

//  return tensorPooling(input_ptr, poolFunction, window_height, window_width,
//		       vertical_pad, horizontal_pad, vertical_stride, horizontal_stride);

  // Only mapped to GPU - get a GPU configuration
  GPUNodeConfiguration *GPUConf =
    (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);

  std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                          std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                 int> > > > &ApproxChoices =
    GPUConf->getApproxChoices();

  // Approximation choices must be for a single operation
  CUSTOM_ASSERT(ApproxChoices.size() == 1 &&
                "Invalid configuration generated for tensor pool wrapper operation");
  enum GPUNodeConfiguration::TENSOR_OP top = ApproxChoices[0].first;
  // Approximation choices must be for a pool operation
  CUSTOM_ASSERT((top == GPUNodeConfiguration::TENSOR_OP::POOL_MAX  ||
                 top == GPUNodeConfiguration::TENSOR_OP::POOL_MEAN ||
                 top == GPUNodeConfiguration::TENSOR_OP::POOL_MIN) &&
         "Invalid configuration generated for tensor pool wrapper operation");

  return handleTensorPoolingApproximationTuples(ApproxChoices[0].second,
                                                input_ptr, poolFunction,
                                                window_height, window_width,
                                                vertical_pad, horizontal_pad,
                                                vertical_stride, horizontal_stride);

}


void* wrapper_tensorGroupConvolution(const char* hpvm_node_id,
                                     void* input, void* filter,
				                     int vertical_pad, int horizontal_pad,
				                     int vertical_stride, int horizontal_stride,
				                     int conv_mode, int conv_groups){
  // Only mapped to GPU - get a GPU configuration
  GPUNodeConfiguration *GPUConf =
    (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);

  std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                          std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                 int> > > > &ApproxChoices =
    GPUConf->getApproxChoices();

  // Approximation choices must be for a group_conv operation
  CUSTOM_ASSERT(ApproxChoices.size() == 1 &&
         ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::GROUP_CONV &&
         "Invalid configuration generated for tensor group_conv wrapper operation");

  return handleTensorGroupConvApproximationTuples(ApproxChoices[0].second,
                                                  input, filter,
                                                  vertical_pad, horizontal_pad,
                                                  vertical_stride, horizontal_stride,
                                                  conv_mode, conv_groups);

}



void* wrapper_tensorSoftmax(const char* hpvm_node_id, void* input_ptr){
//  return tensorSoftmax(input_ptr);

  // Only mapped to GPU - get a GPU configuration
  GPUNodeConfiguration *GPUConf =
    (GPUNodeConfiguration *)RC->getNodeConfiguration(hpvm_node_id);

  std::vector< std::pair< GPUNodeConfiguration::TENSOR_OP,
                          std::vector< std::pair<GPUNodeConfiguration::APPROX,
                                                 int> > > > &ApproxChoices =
    GPUConf->getApproxChoices();

  // Approximation choices must be for a softmax operation
  CUSTOM_ASSERT(ApproxChoices.size() == 1 &&
         ApproxChoices[0].first == GPUNodeConfiguration::TENSOR_OP::SOFTMAX &&
         "Invalid configuration generated for tensor softmax wrapper operation");

  return handleTensorSoftmaxApproximationTuples(ApproxChoices[0].second, input_ptr);


}



